“case-expressions” as a substitute for |with| clauses.

-------------------------------------------------------------------------------

The |with|-clause sometimes causes a problem with the termination
checker, see https://github.com/agda/agda/issues/59.

However, the fix of replacing with by case *doesn't* work.

merge : Leftist ℕ → Leftist ℕ → Leftist ℕ
merge Leaf Leaf = Leaf
merge Leaf x@(Node m t u) = x
merge x@(Node m t u) Leaf = x
merge x₁@(Node m₁ t₁ u₁) x₂@(Node m₂ t₂ u₂) =
  case m₁ ≼-or-≽ m₂ of λ where
    (inl m₁≼m₂) → node m₁ t₁ (merge u₁ x₂)
    (inr m₁≽m₂) → node m₂ t₂ (merge x₁ u₂)

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Various other attempts to replace |with| by a function that performs
the cause analysis *failed*, as well!

infix 0 relate_and_LE:_GE:_
relate_and_LE:_GE:_ : {A : Set} → (a : Carrier) → (b : Carrier) → ({ev : a ⩽ b} → A) → ({ev : a ⩾ b} → A) → A
relate a and b LE: le GE: ge with a ⩽?⩾ b
... | LE a⩽b = le {ev = a⩽b}
... | GE a⩾b = ge {ev = a⩾b}

infix 0 relate_and_LE:_GE:_
relate_and_LE:_GE:_ : {A : Set} → (a : ℕ) → (b : ℕ) → ((ev : a ≼ b) → A) → ((ev : a ≽ b) → A) → A
relate a and b LE: le GE: ge with a ≼?≽ b
... | inl a≼b = le a≼b
... | inr a≽b = ge a≽b
\end{code}

infix 0 relate_and_LE:_GE:_
relate_and_LE:_GE:_ : {A : Set} → (a : ℕ) → (b : ℕ) → (⦃ ev : a ≼ b ⦄ → A) → (⦃ ev : a ≽ b ⦄ → A) → A
relate a and b LE: le GE: ge with a ≼?≽ b
... | inl a≼b = le ⦃ ev = a≼b ⦄
... | inr a≽b = ge ⦃ ev = a≽b ⦄

If we pass additional information into the branches, the problem with
the termination checker re-appears. E.g. merge for skew-heaps (see
VFP.Skew-Heap.Bounded):

_merge_ : Heap≽ n → Heap≽ n → Heap≽ n
Leaf             merge Leaf         = Leaf
Leaf             merge Node a p t u = Node a p t u
Node a p t u     merge Leaf         = Node a p t u
Node a₁ p₁ t₁ u₁ merge Node a₂ p₂ t₂ u₂ =
  relate a₁ and a₂
    LE: Node a₁ p₁ u₁ (t₁ merge Node a₂ it t₂ u₂)
    GE: Node a₂ p₂ u₂ (t₂ merge Node a₁ it t₁ u₁)

The very moment we type |it|, the code turns orange.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Case where

infix 0 case_of_
\end{code}

-------------------------------------------------------------------------------

\begin{code}
case_of_ : {A : Set} {B : Set} → A → (A → B) → B
case x of f = f x
\end{code}

-------------------------------------------------------------------------------

Further reading:

●  https://github.com/agda/agda/issues/59

„“ “” – — → holes: ◌
