Propositional and Leibniz equality, plus equality as the smallest
reflexive relation.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe --cubical-compatible #-}

module VFP.Equality where

open import VFP.Logic

private
  variable
    A : Set
--    R : A → A → Set

infix 6 _≑_  
\end{code}

-------------------------------------------------------------------------------

Propositional equality.

<<propositional-equality>>
\begin{code}
infix 6 _≡_  
data _≡_ (x : A) : A → Set  where
  ≡-reflexive : x ≡ x
\end{code}

\begin{code}
infix 6 _≢_

_≢_ : A → A → Set
a ≢ b = ¬ (a ≡ b)
\end{code}

  instance ≡-reflexive : x ≡ x
 
\begin{code}
pattern reflexive = ≡-reflexive

{-# BUILTIN EQUALITY _≡_  #-}
\end{code}

The pragma enables us to use |rewrite|.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

If we need to remember the result of a function call:

\begin{code}
module Inspect where
  record Reveal_of_is_ {A : Set} {B : A → Set} (f : (x : A) → B x) (x : A) (y : B x) : Set where
    constructor [_]
    field eq : f x ≡ y

  inspect : ∀ {A : Set} {B : A → Set} (f : (x : A) → B x) (x : A) → Reveal f of x is f x
  inspect f x = [ reflexive ]
\end{code}

-------------------------------------------------------------------------------

Leibniz equality states that |a ≐ b| holds if every property |P| that
holds of |a| also holds of |b|.

<<Leibniz-equality>>
\begin{code}
infix 6 _≐_  
_≐_ : A → A → Set₁
a ≐ b = ∀ (P : _ → Set) → P a → P b
\end{code}

-- _≡′_ {A} a b = ∀ (P : A → Set) → P a → P b
--a ≐ b = ∀ P → P a → P b

-------------------------------------------------------------------------------

Defines equality in terms of reflexivity, as the intersection of all
reflexive relations.

<<smallest-reflexive>>
\begin{code}
_≑_ : A → A → Set₁
a ≑ b = ∀ (R : _ → _ → Set) → (∀ {x} → R x x) → R a b
\end{code}

Nice (?) exercise. See also „An Introduction to Mathematical Logic and
Type Theory: To Truth Through Proof“ by Peter P. Andrews, Page 155.

-------------------------------------------------------------------------------

\begin{code}
--Extensionality : Set₁
--Extensionality = {A B : Set} {f g : A → B} → (∀ x → f x ≡ g x) → f ≡ g

Extensionality : Set₁
Extensionality = {A : Set} {B : A → Set} {f g : (x : A) → B x} → (∀ x → f x ≡ g x) → f ≡ g
\end{code}

-------------------------------------------------------------------------------

Unicode table:

≡    \==            U-2261  IDENTICAL TO

-------------------------------------------------------------------------------

Further reading:

●  https://plfa.github.io/Isomorphism/
   see, Section "Extensionality"
●  https://mathoverflow.net/questions/156238/function-extensionality-does-it-make-a-difference-why-would-one-keep-it-out-of
   see, in particular, Andrej Bauer's answer
●  https://github.com/coq/coq/wiki/extensional_equality
