Equivalences of the (strict) orderings.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural.Order.Interrelations where

open import VFP.Natural.Properties hiding (transitive)
open import VFP.Natural.Order
open import VFP.Equality.Properties hiding (transitive) 
open import VFP.Logic.Properties
open import VFP.Decidable

private
  variable
    a a₁ a₂ b b₁ b₂ k m n p : ℕ
\end{code}

-------------------------------------------------------------------------------

<<transitive>>
\begin{code}
≺-transitive : k ≺ m → m ≺ n → k ≺ n
≺-transitive (zero≺succ)     (succ≺succ _  ) = zero≺succ
≺-transitive (succ≺succ k≺m) (succ≺succ m≺n) = succ≺succ (≺-transitive k≺m m≺n)
\end{code}

\begin{code}
≺₇-transitive : k ≺₇ m → m ≺₇ n → k ≺₇ n
≺₇-transitive (one)    q = succ q
≺₇-transitive (succ p) q = succ (≺₇-transitive p q)

zero-bottom₇ : ∀ {n} -> zero ≺₇ succ n
zero-bottom₇ {zero}   = one
zero-bottom₇ {succ n} = ≺₇-transitive zero-bottom₇ one

succ-monotone₇ : ∀ {m n} -> m ≺₇ n → succ m ≺₇ succ n   -- difference is unchanged
succ-monotone₇ (one)    = one
succ-monotone₇ (succ p) = succ (succ-monotone₇ p)
\end{code}

\begin{code}
to-≺₇ : ∀ {m n} -> m ≺ n -> m ≺₇ n
to-≺₇ (zero≺succ)     = zero-bottom₇
to-≺₇ (succ≺succ m≺n) = succ-monotone₇ (to-≺₇ m≺n)

fro-≺₇ : ∀ {m n} -> m ≺₇ n -> m ≺ n
fro-≺₇ (one)    = n≺succ-n
fro-≺₇ (succ p) = ≺-transitive n≺succ-n (fro-≺₇ p)

to-≼₇ : ∀ {m n} -> m ≼ n -> m ≼₇ n
to-≼₇ (inl x) = inl x
to-≼₇ (inr x) = inr (to-≺₇ x)

fro-≼₇ : ∀ {m n} -> m ≼₇ n -> m ≼ n
fro-≼₇ (inl x) = inl x
fro-≼₇ (inr x) = inr (fro-≺₇ x)
\end{code}
