Basic type constructors aka logical connectives.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe --cubical-compatible #-}

module VFP.Logic where

infixr 6 _,_ _&_
infix  5 ¬_
infixr 4 _×_
infixr 3 _⊎_
infixr 2 _←_ _↔_
\end{code}

NB. The precedence of |->| is lower than any precedence you may
declare with |infixl| and |infixr|.

-------------------------------------------------------------------------------

The empty type: absurdity/falsity.

<<falsity>>
\begin{code}
data ⊥ : Set where
\end{code}

<<ex-falso-quodlibet>>
\begin{code}
ex-falso-quodlibet : ∀ {A : Set} -> ⊥ -> A
ex-falso-quodlibet ()
\end{code}

\begin{code}
ex-falso = ex-falso-quodlibet
\end{code}

-------------------------------------------------------------------------------

The unit type: truth.

<<truth>>
\begin{code}
record ⊤ : Set where
  instance constructor trivial
\end{code}

<<truth-non-record>>
\begin{code}
data TT : Set where
  trivial : TT
\end{code}

"instance" is needed to define a Number instance for index sets.

-------------------------------------------------------------------------------

Sums (aka coproducts aka disjoint unions): disjunction.

<<disjoint-union>>
\begin{code}
data _⊎_ (A : Set) (B : Set) : Set where
  inl  : A → A ⊎ B
  inr  : B → A ⊎ B
\end{code}

\begin{code}
case : ∀ {A B C : Set} -> (A → C) → (B → C) → (A ⊎ B → C)
case f g (inl p) = f p
case f g (inr q) = g q
\end{code}

-------------------------------------------------------------------------------

Reintroduce non-dependent pairs? Hmm, this poses problems with let-bindings:
cannot split on argument of unresolved type, see Comparator.

Products (aka pairs): conjunction.
%
begin{code}
record _×_ (A : Set) (B : Set) : Set where
  constructor _,_
  field outl : A
        outr : B
open _×_ public
end{code}

Products (aka pairs): conjunction, defined in terms of dependent
products (aka dependent sums).

<<dependent-product>>
\begin{code}
record Σ (A : Set) (B : A → Set) : Set where
  constructor _,_
  field  outl  : A
         outr  : B outl
open Σ public
\end{code}

<<product>>
\begin{code}
_×_ : ∀ (A : Set) (B : Set) → Set
A × B = Σ A (λ _ -> B)
\end{code}

Functorial action.

\begin{code}
_✕_ : {A A′ B B′ : Set} → (A → A′) → (B → B′) →  (A × B → A′ × B′)
(f ✕ g) (a , b) = f a , g b
\end{code}

NB. The object map is the MULTIPLICATION SIGN #xd7, the arrow map is
the MULTIPLICATION X #x2715.

<<product-simple>>
\begin{code}
data _XX_ (A : Set) (B : Set) : Set where
  mkPair : A → B → A XX B
\end{code}

<<existential>>
\begin{code}
∃ : ∀ {A : Set} → (A → Set) → Set
∃ = Σ _
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Pairs for the next universe level.

\begin{code}
module Next-Universe where
  infixr 4 _×₁_

  record Σ₁ (A : Set₁) (B : A → Set₁) : Set₁ where
    constructor _,_
    field outl : A
          outr : B outl
  open Σ₁ public

  _×₁_ : ∀ (A : Set₁) (B : Set₁) → Set₁
  A ×₁ B = Σ₁ A (λ x -> B)
\end{code}

-------------------------------------------------------------------------------

Negation.

<<negation>>
\begin{code}
¬_ : Set → Set
¬ P = P → ⊥
\end{code}

-------------------------------------------------------------------------------

Implication, reverse implication, and bi-implication aka logical
equivalence.

OR: _⟹_ _⟸_ _⟺_ _iff_

<<reverse-implication>>
\begin{code}
_←_ : Set -> Set -> Set
A ← B = B → A
\end{code}

<<bi-implication>>
\begin{code}
record _↔_ (A B : Set) : Set where
  constructor _&_
  field  to   : A → B
         fro  : B → A
open _↔_ public
\end{code}
