Properties of the strict ordering(s).

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural.Order.Properties-7 where

open import VFP.Natural.Properties hiding (transitive) public
open import VFP.Natural.Order using (one; succ) renaming (_≺_  to _≺₀_;
                                                          _≼_  to _≼₀_;
                                                          _≻_  to _≻₀_;
                                                          _≽_  to _≽₀_;
                                                          _≺₇_ to _≺_;
                                                          _≼₇_ to _≼_;
                                                          _≻₇_ to _≻_;
                                                          _≽₇_ to _≽_)
                                                         public
open import VFP.Natural.Order.Interrelations using (≺₇-transitive;
                                                    zero-bottom₇;
                                                    succ-monotone₇)
                                          renaming (to-≺₇ to to-≺;
                                                    fro-≺₇ to fro-≺;
                                                    to-≼₇ to to-≼;
                                                    fro-≼₇ to fro-≼)
open import VFP.Natural.Order.Properties using () renaming (succ≺succ to succ≺₀succ;
                                                            ≺-irreflexive to ≺₀-irreflexive;
                                                            _≼-or-≻_ to _≼-or-≻₀_;
                                                            _≼-or-≽_ to _≼-or-≽₀_)
open import VFP.Equality.Properties hiding (transitive) public
open import VFP.Logic.Properties
open import VFP.Decidable

private
  variable
    a a₁ a₂ b b₁ b₂ k m n p : ℕ
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₇|.

\begin{code}
≼-reflexive : n ≼ n
≼-reflexive = inl reflexive

≺-transitive : k ≺ m → m ≺ n → k ≺ n
≺-transitive = ≺₇-transitive

≼-transitive : k ≼ m → m ≼ n → k ≼ n
≼-transitive (inl k≡m) m≼n rewrite k≡m = m≼n
≼-transitive k≼m (inl m≡n) rewrite m≡n = k≼m
≼-transitive (inr k≺m) (inr m≺n) = inr (≺-transitive k≺m m≺n)

≺-then-≼ : ∀ {a b c} → a ≺ b → b ≼ c → a ≺ c
≺-then-≼ a≺b (inl reflexive) = a≺b
≺-then-≼ a≺b (inr b≺c)       = ≺-transitive a≺b b≺c

≼-then-≺ : ∀ {a b c} → a ≼ b → b ≺ c → a ≺ c
≼-then-≺ (inl reflexive) b≺c = b≺c
≼-then-≺ (inr a≺b)       b≺c = ≺-transitive a≺b b≺c

zero-bottom : ∀ {n} -> zero ≺ succ n
zero-bottom = zero-bottom₇

zero≺succ : ∀ {n} -> zero ≺ succ n
zero≺succ = zero-bottom

zero≺succ-n : ∀ n -> zero ≺ succ n
zero≺succ-n n = zero-bottom {n = n}

zero≼n : ∀ n → zero ≼ n
zero≼n (zero)   = inl reflexive 
zero≼n (succ _) = inr zero≺succ

n≼succ-n : ∀ {n} → n ≼ succ n
n≼succ-n = inr one

n≺succ-n : ∀ {n} → n ≺ succ n
n≺succ-n = one

succ-monotone : ∀ {m n} -> m ≺ n → succ m ≺ succ n
succ-monotone = succ-monotone₇
\end{code}


\begin{code}
succ-strict-order-reflecting : succ m ≺ succ n -> m ≺ n
succ-strict-order-reflecting one = one
succ-strict-order-reflecting (succ p) = succ (succ-strict-order-reflecting p)

succ-order-reflecting : succ m ≼ succ n → m ≼ n
succ-order-reflecting (inl reflexive) = inl reflexive
succ-order-reflecting (inr p) = inr (succ-strict-order-reflecting p)

succ-reflects-≺ = succ-strict-order-reflecting

succ-≼-is-≺ : succ m ≼ n -> m ≺ n
succ-≼-is-≺ (inl reflexive) = one
succ-≼-is-≺ (inr y)         = ≺-transitive one y

≺-is-succ-≼ : m ≺ n -> succ m ≼ n 
≺-is-succ-≼ one = inl reflexive
≺-is-succ-≼ (succ p) = inr p

≼-is-≺-succ : m ≼ n -> m ≺ succ n
≼-is-≺-succ (inl reflexive) = one
≼-is-≺-succ (inr y)         = ≺-transitive y one

≺-succ-is-≼ : m ≺ succ n -> m ≼ n
≺-succ-is-≼ one = inl reflexive
≺-succ-is-≼ (succ p) = inr (succ-reflects-≺ p)

succ≺succ : m ≺ n -> succ m ≺ succ n
succ≺succ p = to-≺ (succ≺₀succ (fro-≺ p))

succ≼succ : m ≼ n -> succ m ≼ succ n
succ≼succ (inl reflexive) = inl reflexive
succ≼succ (inr p) = inr (succ≺succ p)

¬n≺zero : ¬ n ≺ zero
¬n≺zero (succ p) = ¬n≺zero p

≺-irreflexive : ¬ n ≺ n
≺-irreflexive p = ≺₀-irreflexive (fro-≺ p)

not-greater : ∀ {m n} -> m ≺ n -> ¬ m ≻ n
not-greater a≺b b≺a = ≺-irreflexive (≺-transitive a≺b b≺a)

≼-antisymmetric : m ≼ n → n ≼ m → m ≡ n
≼-antisymmetric (inl reflexive) n≼m             = reflexive
≼-antisymmetric (inr m≺n)       (inl reflexive) = reflexive
≼-antisymmetric (inr m≺n)       (inr n≺m)       = ex-falso-quodlibet (not-greater m≺n n≺m)

the-one : m ≼ n -> n ≺ 1 + m -> m ≡ n
the-one p q = ≼-antisymmetric p (≺-succ-is-≼ q)
\end{code}

\begin{code}
≺-monotone-left-+ : ∀ {k m n} -> m ≺ n -> m + k ≺ n + k
≺-monotone-left-+ one = one
≺-monotone-left-+ (succ p) = succ (≺-monotone-left-+ p)

≺-monotone-right-+ : ∀ {k m n} -> m ≺ n -> k + m ≺ k + n
≺-monotone-right-+ {k = zero} p = p
≺-monotone-right-+ {k = succ k} p = succ≺succ (≺-monotone-right-+ p)

≼-monotone-left-+ : ∀ {k m n} -> m ≼ n -> m + k ≼ n + k
≼-monotone-left-+ (inl reflexive) = inl reflexive
≼-monotone-left-+ (inr y) = inr (≺-monotone-left-+ y)

≼-monotone-right-+ : ∀ {k m n} -> m ≼ n -> k + m ≼ k + n
≼-monotone-right-+ (inl reflexive) = ≼-reflexive
≼-monotone-right-+ (inr y) = inr (≺-monotone-right-+ y)
\end{code}

Monus
\begin{code}
m∸n≼m : ∀ {m n} → m ∸ n ≼ m
m∸n≼m {zero} {n} = inl reflexive
m∸n≼m {succ m} {zero} = inl reflexive
m∸n≼m {succ m} {succ n} = ≼-transitive m∸n≼m n≼succ-n

∸-shift : ∀ {m n} → n ≼ m → succ m ∸ n ≡ succ (m ∸ n)
∸-shift {zero} {zero} p = reflexive
∸-shift {zero} {succ n} p = ex-falso-quodlibet (≺-irreflexive (≼-then-≺ p (zero≺succ-n n)))
∸-shift {succ m} {zero} p = reflexive
∸-shift {succ m} {succ n} p = ∸-shift (succ-order-reflecting p)

∸-positive : ∀ {m n} -> m ≺ n -> zero ≺ n ∸ m
∸-positive {m} {zero} m≺n = ≼-then-≺ (zero≼n m) m≺n
∸-positive {zero} {succ n} m≺n = m≺n
∸-positive {succ m} {succ n} m≺n = ∸-positive (succ-strict-order-reflecting m≺n)

∸-pseudo-monotone : ∀ {a b n} → n ≺ b → a ≺ b → a ∸ n ≺ b ∸ n
∸-pseudo-monotone {zero} {b} {n} n≺b a≺b = ∸-positive n≺b
∸-pseudo-monotone {succ a} {b} {zero} n≺b a≺b rewrite ∸-right-identity {b} = a≺b
∸-pseudo-monotone {succ a} {zero} {succ n} n≺b a≺b = ex-falso-quodlibet (≺-irreflexive (≺-transitive a≺b zero≺succ))
∸-pseudo-monotone {succ a} {succ b} {succ n} n≺b a≺b = ∸-pseudo-monotone (succ-strict-order-reflecting n≺b) (succ-strict-order-reflecting a≺b)

∸-pseudo-antitone : ∀ {a b n} -> b ≺ n -> b ≺ a -> n ∸ a ≺ n ∸ b
∸-pseudo-antitone {a} {b} {zero} b≺n b≺a = ≼-then-≺ (zero≼n b) b≺n
∸-pseudo-antitone {zero} {b} {succ n} b≺n b≺a = ex-falso-quodlibet (≺-irreflexive (≼-then-≺ (zero≼n b) b≺a))
∸-pseudo-antitone {succ a} {zero} {succ n} b≺n b≺a = ≼-is-≺-succ m∸n≼m
∸-pseudo-antitone {succ a} {succ b} {succ n} b≺n b≺a = ∸-pseudo-antitone (succ-strict-order-reflecting b≺n) (succ-strict-order-reflecting b≺a)
\end{code}


Accessibility.

<<accessibility>>
\begin{code}
data Accessible (n : ℕ) : Set where
  access : (∀ {m} → m ≺ n → Accessible m) → Accessible n
\end{code}

<<accessibility-zero>>
\begin{code}
old-acc-zero : Accessible zero
old-acc-zero = access (λ m≺0 → ex-falso-quodlibet (¬n≺zero m≺0))
\end{code}

<<accessibility-step-signature>>
\begin{code}
old-acc-succ : ∀ {n : ℕ} -> Accessible n → Accessible (succ n)
old-acc-succ {n} (access hyp) = access hyp′
  where hyp′ : {m : ℕ} → m ≺ succ n → Accessible m
\end{code}

<<accessibility-assumption>>
\begin{code}
        hyp′ one = access hyp
\end{code}
<<accessibility-strictly-smaller>>
\begin{code}
        hyp′ (succ succ-m≺succ-n) = hyp (succ-reflects-≺ succ-m≺succ-n)
\end{code}

\begin{code}
old-well-founded : ∀ n → Accessible n
old-well-founded (zero)    = old-acc-zero
old-well-founded (succ n)  = old-acc-succ (old-well-founded n)
\end{code}

RH: some experiments.

<<alt-accessibility-zero>>
\begin{code}
acc-zero : m ≺ zero → Accessible m
acc-zero m≺0 = ex-falso-quodlibet (¬n≺zero m≺0)
\end{code}

<<alt-accessibility-succ-signature>>
\begin{code}
acc-succ : Accessible n → m ≺ succ n → Accessible m
\end{code}
<<include src/Natural/Order/Properties-alt-accessibility-7-well-founded.snip>>

<<alt-accessibility-succ-assumption>>
\begin{code}
acc-succ (access because) (one)                 = access because
\end{code}

<<alt-accessibility-succ-strictly-smaller>>
\begin{code}
acc-succ (access because) (succ succ-m≺succ-n)  = because (succ-reflects-≺ succ-m≺succ-n)
\end{code}

<<alt-accessibility-well-founded-sig>>
\begin{code}
well-founded : ∀ n → Accessible n
\end{code}

<<alt-accessibility-well-founded-impl>>
\begin{code}
well-founded (zero)    = access acc-zero
well-founded (succ n)  = access (acc-succ (well-founded n))
\end{code}

-----

Same as above, only different names.

<<access-zero>>
\begin{code}
access-zero : m ≺ 0 → Accessible m
access-zero m≺0 = ex-falso-quodlibet (¬n≺zero m≺0)
\end{code}

<<access-succ>>
\begin{code}
access-succ : Accessible n → m ≺ 1 + n → Accessible m
access-succ (access because) (one)                 = access because
access-succ (access because) (succ succ-m≺succ-n)  = 
  because (succ-reflects-≺ succ-m≺succ-n)
\end{code}

<<accessible>>
\begin{code}
accessible : ∀ {n} → Accessible n
accessible {zero}    = access access-zero
accessible {succ n}  = access (access-succ (accessible {n}))
\end{code}

\begin{code}
infix 6 _≼-or-≻_ _≺-or-≽_
_≼-or-≻_ : ∀ m n -> m ≼ n ⊎ m ≻ n
m ≼-or-≻ n with m ≼-or-≻₀ n
... | inl m≼n = inl (to-≼ m≼n)
... | inr m≻n = inr (to-≺ m≻n)

_≺-or-≽_ : ∀ m n -> m ≺ n ⊎ m ≽ n
m ≺-or-≽ n with n ≼-or-≻ m
... | inl n≼m = inr n≼m
... | inr m≻n = inl m≻n
\end{code}

\begin{code}
infix 6 _≼-or-≽_
_≼-or-≽_ : ∀ m n -> m ≼ n ⊎ m ≽ n
m ≼-or-≽ n with m ≼-or-≽₀ n
... | inl m≼n = inl (to-≼ m≼n)
... | inr m≽n = inr (to-≼ m≽n)
\end{code}

\begin{code}
infix 6 _≼?≻_ _≼?≽_ _≺?≽_
_≼?≻_ = _≼-or-≻_
_≼?≽_ = _≼-or-≽_
_≺?≽_ = _≺-or-≽_
\end{code}

