Record types for injective, surjective and bijective functions.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe --cubical-compatible #-}

module VFP.Bijection where

open import VFP.Equality public
open import VFP.Logic public

infix  2 _↣_ _↠_ _⤖_ _↪_ _≅_
\end{code}

-------------------------------------------------------------------------------

Injective functions.

\begin{code}
record _↣_ (A B : Set) : Set where
  field to        : A -> B
        injective : ∀ x₁ x₂ -> to x₁ ≡ to x₂ → x₁ ≡ x₂
open _↣_ public
\end{code}

Surjective functions.

\begin{code}
record _↠_ (A B : Set) : Set where
  field to         : A -> B
        surjective : ∀ y -> ∃ (λ x → to x ≡ y)
open _↠_ public
\end{code}

Bijective functions.

\begin{code}
record _⤖_ (A B : Set) : Set where
  field to         : A -> B
        injective  : ∀ x₁ x₂ -> to x₁ ≡ to x₂ → x₁ ≡ x₂
        surjective : ∀ y -> ∃ (λ x → to x ≡ y)
open _⤖_ public
\end{code}

Left-invertable functions.

\begin{code}
record _↪_ (A B : Set) : Set where
  field to  : A -> B
        fro : B -> A
        fro∘to≡id : ∀ x -> fro (to x) ≡ x
open _↪_ public
\end{code}

A function with a *nonempty* domain is injective if and only if it is
left-invertable i.e. it has a left inverse.

Bijections or one-to-one mappings (isomorphisms) - alternative
definition.

\begin{code}
record _≅_ (A B : Set) : Set where
  field to  : A -> B
        fro : B -> A
        fro∘to≡id : ∀ x -> fro (to x) ≡ x
        to∘fro≡id : ∀ x -> to (fro x) ≡ x
open _≅_ public
\end{code}

A function is bijective if and only if it is both left-invertable and
right-invertable.

-------------------------------------------------------------------------------

Unicode table:

↪    \hookrightarrow    U-21aa  RIGHTWARDS ARROW WITH HOOK
≅    \cong \~=          U-2245  APPROXIMATELY EQUAL TO
∘    \comp or \circ or \o
