\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Integer where

import VFP.Natural as ℕ
open ℕ using (ℕ)

import VFP.Natural.Properties as ℕ-Props
import VFP.Natural.Order as ℕ-Ord
import VFP.Natural.Order.Properties as ℕ-OrdProps

open import VFP.Logic
open import VFP.Equality
open import VFP.Equality.Properties
open import VFP.Equality.Reasoning

\end{code}

\begin{code}
infix 8 +_  -- Standard library uses this as +_

data ℤ : Set where
  +_    : (n : ℕ) → ℤ
  -[1+_] : (n : ℕ) → ℤ

{-# BUILTIN INTEGER       ℤ    #-}
{-# BUILTIN INTEGERPOS    +_    #-}
{-# BUILTIN INTEGERNEGSUC -[1+_] #-}

--primitive primShowInteger : Int → String
\end{code}

\begin{code}
succ : ℤ → ℤ
succ (+ n) = + ℕ.succ n
succ -[1+ ℕ-Props.zero ] = + ℕ.zero
succ -[1+ ℕ-Props.succ n ] = -[1+ n ]

pred : ℤ → ℤ
pred (+ ℕ.zero) = -[1+ ℕ.zero ]
pred (+ ℕ.succ n) = + n
pred -[1+ n ] = -[1+ ℕ.succ n ]
\end{code}

\begin{code}
pattern +0       = + 0
pattern +[1+_] n = + (ℕ.succ n)
\end{code}

\begin{code}
0ℤ : ℤ
0ℤ = +0
\end{code}

\begin{code}
∣_∣ : ℤ → ℕ
∣ + n      ∣ = n
∣ -[1+ n ] ∣ = ℕ.succ n

-_ : ℤ → ℤ
- -[1+ n ] = +[1+ n ]
- +0       = +0
- +[1+ n ] = -[1+ n ]

_⊖_ : ℕ → ℕ → ℤ
m ⊖ ℕ-Props.zero = + m
m ⊖ ℕ-Props.succ n = pred (m ⊖ n)

-- Addition.

_+_ : ℤ → ℤ → ℤ
-[1+ m ]  + -[1+ n ] = -[1+ ℕ.succ (m ℕ.+ n) ]
-[1+ m ]  + (+    n) = n ⊖ ℕ.succ m
(+    m)  + -[1+ n ] = m ⊖ ℕ.succ n
(+    m ) + (+   n)  = + (m ℕ.+ n)

-- -- Subtraction.

_-_ : ℤ → ℤ → ℤ
i - j = i + (- j)
\end{code}
