\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Integer.Properties where

import VFP.Natural as ℕ
open ℕ using (ℕ)

import VFP.Natural.Properties as ℕ-Props
import VFP.Natural.Order as ℕ-Ord
import VFP.Natural.Order.Properties as ℕ-OrdProps

open import VFP.Logic
open import VFP.Equality
open import VFP.Equality.Properties
open import VFP.Equality.Reasoning

open import VFP.Integer

\end{code}

\begin{code}
succ-pred-cancel : {n : ℤ} → n ≡ succ (pred n)
succ-pred-cancel {+ ℕ-Props.zero} = reflexive
succ-pred-cancel {+ ℕ-Props.succ n} = reflexive
succ-pred-cancel { -[1+ n ]} = reflexive

pred-succ-cancel : {n : ℤ} → n ≡ pred (succ n)
pred-succ-cancel {+ n} = reflexive
pred-succ-cancel { -[1+ ℕ-OrdProps.zero ]} = reflexive
pred-succ-cancel { -[1+ ℕ-OrdProps.succ n ]} = reflexive
\end{code}

\begin{code}
pred-n⊖m≡n⊖succ-m : {n m : ℕ} → pred (n ⊖ m) ≡ n ⊖ (ℕ.succ m)
pred-n⊖m≡n⊖succ-m = reflexive

pred-0⊖n≡-[1+n] : {n : ℕ} → pred (0 ⊖ n) ≡ -[1+ n ]
pred-0⊖n≡-[1+n] {ℕ-OrdProps.zero} = reflexive
pred-0⊖n≡-[1+n] {ℕ-OrdProps.succ n} = congruent pred pred-0⊖n≡-[1+n]

+-left-identity : {n : ℤ} → 0ℤ + n ≡ n
+-left-identity {+ n} = reflexive
+-left-identity { -[1+ ℕ-OrdProps.zero ]} = reflexive
+-left-identity { -[1+ ℕ-OrdProps.succ n ]} = congruent pred (+-left-identity { -[1+ n ] })

+-right-identity : {n : ℤ} → n + 0ℤ ≡ n
+-right-identity {+ ℕ-OrdProps.zero} = reflexive
+-right-identity {+[1+ n ]} = congruent succ (+-right-identity {+ n})
+-right-identity { -[1+ n ]} = pred-0⊖n≡-[1+n]

+-commutative : {m n : ℤ} → m + n ≡ n + m
+-commutative {+ n} {+ m} = congruent (λ x → + x) (ℕ-Props.+-commutative n m)
+-commutative {+ n} { -[1+ m ]} = reflexive
+-commutative { -[1+ n ]} {+ m} = reflexive
+-commutative { -[1+ n ]} { -[1+ m ]} = congruent (λ x → -[1+ ℕ.succ x ]) (ℕ-Props.+-commutative n m)

-- +-associative : {l m n : ℤ} → (l + m) + n ≡ l + (m + n)
-- +-associative {+ l} {+ m} {+ n} = congruent (λ x → + x) (ℕ-Props.+-associative l)
-- +-associative {+ l} {+ m} { -[1+ n ]} = {!!}
-- +-associative {+ l} { -[1+ m ]} {+ n} = {!!}
-- +-associative {+ l} { -[1+ m ]} { -[1+ n ]} = {!!}
-- +-associative { -[1+ l ]} {+ m} {+ n} = {!!}
-- +-associative { -[1+ l ]} {+ m} { -[1+ n ]} = {!!}
-- +-associative { -[1+ l ]} { -[1+ m ]} {+ n} = {!!}
-- +-associative { -[1+ l ]} { -[1+ m ]} { -[1+ n ]} = {!!}
\end{code}

\begin{code}
⟨succ-n⊖m⟩≡succ-⟨n⊖m⟩ : {n m : ℕ} → (ℕ.succ n ⊖ m) ≡ succ (n ⊖ m)
⟨succ-n⊖m⟩≡succ-⟨n⊖m⟩ {n} {ℕ-OrdProps.zero} = reflexive
⟨succ-n⊖m⟩≡succ-⟨n⊖m⟩ {n} {ℕ-OrdProps.succ m} = ≡-transitive (congruent pred (⟨succ-n⊖m⟩≡succ-⟨n⊖m⟩ {n} {m})) (≡-transitive (≡-symmetric pred-succ-cancel) succ-pred-cancel)

pred-⟨succ-⟨n⊖m⟩⟩≡n⊖m : {n m : ℕ} → pred (ℕ.succ n ⊖ m) ≡ n ⊖ m
pred-⟨succ-⟨n⊖m⟩⟩≡n⊖m {n} {m} =
  pred (ℕ-OrdProps.succ n ⊖ m)
  ≡⟨ congruent pred (⟨succ-n⊖m⟩≡succ-⟨n⊖m⟩ {n} {m} )⟩
  pred (succ (n ⊖ m))
  ≡⟨ ≡-symmetric pred-succ-cancel ⟩
  (n ⊖ m)
  ∎
\end{code}

\begin{code}
data even : ℤ → Set where
  even+ : {n : ℕ} → ℕ-Props.even n → even (+ n)
  even- : {n : ℕ} → ℕ-Props.odd n → even -[1+ n ]

odd : ℤ → Set
odd n = even (succ n)
\end{code}

\begin{code}
odd-2 : {n : ℕ} → ℕ-Props.odd (ℕ.succ (ℕ.succ n)) → ℕ-Props.odd n
odd-2 {ℕ-Props.zero} (ℕ-Props.step p) = p
odd-2 {ℕ-Props.succ n} (ℕ-Props.step p) = p

odd-via-pred : {n : ℤ} → even (pred n) → odd n
odd-via-pred {+ ℕ-Props.zero} (even- x) = even+ x
odd-via-pred {+ ℕ-Props.succ n} (even+ x) = even+ (ℕ-Props.step x)
odd-via-pred { -[1+ ℕ-Props.zero ]} (even- odd-succ-n) = even+ ℕ-Props.base
odd-via-pred { -[1+ ℕ-Props.succ n ]} (even- odd-succ-n) = even- (odd-2 odd-succ-n)
\end{code}

\begin{code}
parity : (n : ℤ) → even n ⊎ odd n
parity (+ n) with ℕ-Props.parity n
... | inl x = inl (even+ x)
... | inr x = inr (even+ x)
parity -[1+ n ] with ℕ-Props.parity n
... | inl x = inr (even-succ-[1+n] x)
  where
    even-succ-[1+n] : {n : ℕ} → ℕ-Props.even n → even (succ -[1+ n ])
    even-succ-[1+n] {ℕ-Props.zero} p = even+ p
    even-succ-[1+n] {ℕ-Props.succ n} p = even- p
... | inr x = inl (even- x)
\end{code}
