Properties of logical connectives. A fairly random collection of
(non-) properties.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Logic.Properties where

open import VFP.Logic public

private
  variable
    A B B₁ B₂ C : Set
    X Y Z : Set

infix 0 case_of_
\end{code}

-------------------------------------------------------------------------------

Implication.

\begin{code}
→-reflexive : A → A
→-reflexive = λ x -> x

→-transitive : (A → B) -> (B → C) -> (A → C)
→-transitive AB BC = λ x -> BC (AB x)
\end{code}

Equivalence.

\begin{code}
↔-reflexive : A ↔ A
↔-reflexive = (λ x -> x) & (λ x -> x)

↔-symmetric : A ↔ B -> B ↔ A
↔-symmetric AB = fro AB & to AB

↔-transitive : A ↔ B -> B ↔ C -> A ↔ C
↔-transitive AB BC = (λ x -> to BC (to AB x)) & (λ x -> fro AB (fro BC x))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
×-commutative : (A × B) ↔ (B × A)
×-commutative = (λ (a , b) → (b , a)) & (λ (b , a) → (a , b))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Disjunction aka disjoint union.

\begin{code}
⊎-preserves-↔-right : B₁ ↔ B₂ → (A ⊎ B₁ ↔ A ⊎ B₂)
⊎-preserves-↔-right (to & fro) = (λ where
                                    (inl a)  → inl a
                                    (inr b₁) → inr (to b₁))
                               & (λ where
                                    (inl a)  → inl a
                                    (inr b₂) → inr (fro b₂))

⊎-right-unit : A ⊎ ⊥ ↔ A
⊎-right-unit = (λ where (inl x) → x) & inl
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<distributivity>>
\begin{code}
×-left-distributes-over-⊎ : (A ⊎ B) × C ↔ A × C ⊎ B × C
×-left-distributes-over-⊎ = (λ where
                               (inl a , c) → inl (a , c)
                               (inr b , c) → inr (b , c))
                          & (λ where
                               (inl (a , c)) → inl a , c
                               (inr (b , c)) → inr b , c)
\end{code}

-------------------------------------------------------------------------------

Interdefinability.

<<interdefinability>>
\begin{code}
conjunction-as-disjunction : A × B → ¬ (¬ A ⊎ ¬ B)
conjunction-as-disjunction (a , b) (inl ¬a) = ¬a a
conjunction-as-disjunction (a , b) (inr ¬b) = ¬b b

-- conjunction-as-disjunctionᴿ : A × B ← ¬ (¬ A ⊎ ¬ B)  -- good luck!

disjunction-as-conjunction : A ⊎ B → ¬ (¬ A × ¬ B)
disjunction-as-conjunction (inl a) (¬a , ¬b) = ¬a a
disjunction-as-conjunction (inr b) (¬a , ¬b) = ¬b b

-- disjunction-as-conjunctionᴿ : A ⊎ B ← ¬ (¬ A × ¬ B)  -- good luck!

-- implication-as-disjunction : (A → B) → ¬ A ⊎ B  -- good luck!
\end{code}

-------------------------------------------------------------------------------

<<contraposition>>
\begin{code}
contrapositive : (A → B) → (¬ B → ¬ A)
contrapositive f ¬b = λ a → ¬b (f a)  -- composition

-- contrapositiveᴿ : (A → B) ← (¬ B → ¬ A)  -- good luck!

flip : (A -> ¬ B) -> (B -> ¬ A)
flip f b a = f a b
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

The classical law A ⊎ ¬ A, tertium non datur, (latin for „ein Drittes
gibt es nicht“, law of excluded middle), does *not* hold
intuitionistically. It is, however, irrefutable.

<<excluded-middle-irrefutable>>
\begin{code}
-- excluded-middle : A ⊎ ¬ A  -- good luck!

excluded-middle-irrefutable : ¬ ¬ (A ⊎ ¬ A)
excluded-middle-irrefutable =
  λ neg-excluded-middle → neg-excluded-middle (inr (λ a → neg-excluded-middle (inl a)))
\end{code}

<<excluded-middle-irrefutable-var>>
\begin{code}
excluded-middle-irrefutable′ : ¬ ¬ (A ⊎ ¬ A)
excluded-middle-irrefutable′ ¬-excluded-middle =
  ¬-excluded-middle (inr (λ a → ¬-excluded-middle (inl a)))
\end{code}

\begin{code}
simpler-version : ¬ (A ⊎ ¬ A) -> A ⊎ ¬ A
simpler-version = λ neg-excluded-middle → inr (λ a → neg-excluded-middle (inl a))
\end{code}

We show at the end:

double-negation-irrefutable : ¬ ¬ (¬ ¬ A -> A)

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

De Morgan's laws and co. In classicial logic they can be used to turn
a formula into negation normalform. 

TODO: rename: negation-X to negate-X.

<<single-negation>>
\begin{code}
negation-⊥ : ¬ ⊥ ↔ ⊤
negation-⊥ = (λ p → trivial) & (λ p x → x)

negation-⊤ : ¬ ⊤ ↔ ⊥
negation-⊤ = (λ z → z trivial) & (λ z x → z) 

-- negation-¬ : ¬ ¬ A -> A  -- good luck!
negation-¬ᴿ : ¬ ¬ A ← A
negation-¬ᴿ a ¬a = ¬a a  -- continuations at work

negate-⊎ : ¬ (A ⊎ B) ↔ ¬ A × ¬ B
negate-⊎ = (λ f → (λ a → f (inl a)) , (λ b → f (inr b)))
         & (λ where (¬a , ¬b) (inl a) → ¬a a
                    (¬a , ¬b) (inr b) → ¬b b)

-- negation-× : ¬ (A × B) -> ¬ A ⊎ ¬ B  -- good luck!
negation-×ᴿ : ¬ (A × B) ← ¬ A ⊎ ¬ B
negation-×ᴿ (inl a) = λ p → a (outl p)
negation-×ᴿ (inr b) = λ p → b (outr p)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

The classical law ¬ ¬ A -> A, duplex negatio affirmat (latin „die
doppelte Verneinung bekräftigt/bejaht“, logical double negation),
does *not* hold intuitionistically.

<<double-negation>>
\begin{code}
-- double-negation : ¬ ¬ A -> A  -- good luck!

double-negationᴿ : ¬ ¬ A ← A  -- same as negation-¬ᴿ
double-negationᴿ a ¬a = ¬a a

triple-negation : ¬ ¬ ¬ A -> ¬ A
triple-negation = λ ¬¬¬a a → ¬¬¬a (λ ¬a → ¬a a)

triple-negationᴿ : ¬ ¬ ¬ A ← ¬ A
triple-negationᴿ = double-negationᴿ
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Gentzen's double negation translation, see also
VFP.Logic.Propositional.Natural-Deduction.Double-Negation.

<<double-negation-translation>>
\begin{code}
double-negation-⊥ : ¬ ¬ ⊥ -> ⊥
double-negation-⊥ = λ c → c (λ x → x)

double-negation-⊥ᴿ : ¬ ¬ ⊥ ← ⊥
double-negation-⊥ᴿ = λ z z₁ → z₁ z

double-negation-⊤ : ¬ ¬ ⊤ -> ⊤
double-negation-⊤ = λ x → trivial

double-negation-⊤ᴿ : ¬ ¬ ⊤ ← ⊤
double-negation-⊤ᴿ = λ _ z → z trivial

-- double-negation-⊎ : ¬ ¬ (A ⊎ B) → ¬ ¬ A ⊎ ¬ ¬ B  -- good luck!
double-negation-⊎ᴿ : ¬ ¬ (A ⊎ B) ← ¬ ¬ A ⊎ ¬ ¬ B
double-negation-⊎ᴿ (inl x) = λ z → x (λ z₁ → z (inl z₁))
double-negation-⊎ᴿ (inr x) = λ z → x (λ z₁ → z (inr z₁))

double-negation-var-⊎ : ¬ ¬ (A ⊎ B) → ¬ (¬ ¬ ¬ A × ¬ ¬ ¬ B)
double-negation-var-⊎ f (¬¬¬a , ¬¬¬b) = f (λ where
                                         (inl a) → ¬¬¬a (λ ¬a → ¬a a)
                                         (inr b) → ¬¬¬b (λ ¬b → ¬b b))
double-negation-var-⊎ᴿ : ¬ ¬ (A ⊎ B) ← ¬ (¬ ¬ ¬ A × ¬ ¬ ¬ B)
double-negation-var-⊎ᴿ = λ z z₁ → z ((λ x → x (λ x₁ → z₁ (inl x₁))) , (λ x → x (λ x₁ → z₁ (inr x₁))))

double-negation-× : ¬ ¬ (A × B) → (¬ ¬ A × ¬ ¬ B)
double-negation-× f = (λ ¬a → f (λ (a , b) → ¬a a)) , (λ ¬b → f (λ (a , b) → ¬b b))

double-negation-×ᴿ : ¬ ¬ (A × B) ← (¬ ¬ A × ¬ ¬ B)
double-negation-×ᴿ (f , g) h = f (λ a → g (λ b → h (a , b)))

double-negation-→ : ¬ ¬ (A → B) -> (¬ ¬ A → ¬ ¬ B)
double-negation-→ ¬¬f = λ ¬¬a ¬b → ¬¬a (λ a → ¬¬f (λ f → ¬b (f a)))

double-negation-→ᴿ : ¬ ¬ (A → B) ← (¬ ¬ A → ¬ ¬ B)
double-negation-→ᴿ f = λ x → x (λ a → ex-falso-quodlibet (f (λ c → c a) (λ b → x (λ _ → b))))
\end{code}

NB. A formula is called *stable* if double negation elimination holds for it.


\begin{code}
¬¬-⊎-elim : ¬ ¬ (A ⊎ B) → (A → C) → (B → C) -> ¬ ¬ C
¬¬-⊎-elim f g h k = f (λ where
                         (inl a) → k (g a)
                         (inr b) → k (h b))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

TODO: Examples of double negation translation.

NB. excluded-middle-irrefutable is actually *not* the double negation of
excluded middle, as A is universally quantified: ¬ ¬ (∀ {A : Set} → A ⊎ ¬ A).

¬₁_ : Set₁ → Set₁
¬₁ P = P → ⊥

double-negation-dn : ¬₁ ¬₁ (∀ {A : Set} → ¬ ¬ A -> A)
double-negation-dn = {!!}

double-negation-em : ¬₁ ¬₁ (∀ (A : Set) → A ⊎ ¬ A)
double-negation-em = λ negation-excluded-middle → negation-excluded-middle (λ A -> inr (λ (a : A) → negation-excluded-middle ({!a!})))

-------------------------------------------------------------------------------

\begin{code}
excluded-middle→double-negation-elimination : (A ⊎ ¬ A) -> (¬ ¬ A → A)
excluded-middle→double-negation-elimination (inl  a) c = a
excluded-middle→double-negation-elimination (inr ¬a) c = ex-falso-quodlibet (c ¬a)

¬¬-monotone : (A -> B) -> (¬ ¬ A -> ¬ ¬ B)
¬¬-monotone f = λ ¬¬a ¬b → ¬¬a (λ a → ¬b (f a))

double-negation-irrefutable : ¬ ¬ (¬ ¬ A -> A)
double-negation-irrefutable = ¬¬-monotone excluded-middle→double-negation-elimination excluded-middle-irrefutable
\end{code}

Gentzen's “Erstes Beispiel” (p.183):

<<Gentzen>>
\begin{code}
Distributive : A ⊎ (B × C) -> (A ⊎ B) × (A ⊎ C)
Distributive (inl  a)        = inl  a  , inl  a
Distributive (inr  (b , c))  = inr  b  , inr  c
\end{code}

case_of_ : {A : Set} {B : Set} → A → (A → B) → B

<<Gentzen-var>>
\begin{code}
case_of_ : A → (A → B) → B
case x of f = f x

Distributive′ : A ⊎ (B × C) -> (A ⊎ B) × (A ⊎ C)
Distributive′ = λ p →  case p of λ where
                       (inl  a)  → inl  a         , inl  a
                       (inr  q)  → inr  (outl q)  , inr  (outr q)
\end{code}

-------------------------------------------------------------------------------

COPIED VERBATIM FROM VFP.Logic.Reasoning so that
VFP.Sorting.By-Insertion.Generic compiles. THERE MUST BE A BETTER WAY.

\begin{code}
module ↔-Reasoning where
  private
    data Order (A : Set) (B : Set) : Set where
      LE : (A → B) -> Order A B
      EQ : (A ↔ B) -> Order A B

  module _ {A B : Set} where
    ⟦_⟧ : Order A B -> Set
    ⟦ LE  _ ⟧ = A → B
    ⟦ EQ  _ ⟧ = A ↔ B

    infix 0 proof_

    proof_ : (p : Order A B) → ⟦ p ⟧
    proof (LE p) = p
    proof (EQ p) = p

  infixr 1 →-link ↔-link
  infix  2 _∎
  syntax →-link x q p = x →⟨ p ⟩ q
  syntax ↔-link x q p = x ↔⟨ p ⟩ q

  →-link : (A : Set) {B C : Set} → Order B C → (A → B) → Order A C
  →-link x (LE g) f = LE (→-transitive f g)
  →-link x (EQ g) f = LE (λ x -> to g (f x))

  ↔-link : (A : Set) {B C : Set} → Order B C → (A ↔ B) → Order A C
  ↔-link x (LE g) f = LE (λ x → g (to f x))
  ↔-link x (EQ g) f = EQ (↔-transitive f g)

  _∎ : (A : Set) → Order A A
  x ∎ = EQ ↔-reflexive
\end{code}

-------------------------------------------------------------------------------

Further reading:

●  https://agda.readthedocs.io
●  https://ncatlab.org/nlab/show/double+negation+translation
●  https://math.stackexchange.com/questions/658262/what-is-the-correct-reading-of-bot??
