Optional values (Maybe) aka the partiality monad.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}
module VFP.Maybe where

open import VFP.Equality

private
    variable
        A B : Set
\end{code}

-------------------------------------------------------------------------------

<<Maybe>>
\begin{code}
data Maybe (Elem : Set) : Set where
  Nothing  : Maybe Elem
  Just     : Elem -> Maybe Elem

undefined : Maybe A
undefined = Nothing
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

\begin{code}
map : (A -> B) -> (Maybe A -> Maybe B)
map f Nothing  = Nothing
map f (Just x) = Just (f x)
\end{code}

Proof of the functor laws. Todo:

begin{code}
map-preserves-identity : map identity m ≡ m
map-preserves-identity 
end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

The monad operations.

<<monad>>
\begin{code}
return : A -> Maybe A
return a = Just a

_>>=_ : Maybe A -> (A -> Maybe B) -> Maybe B
Nothing  >>= k = Nothing
Just a   >>= k = k a
\end{code}

Proof of the monad laws.

\begin{code}
private
    variable
        C : Set
        a : A
        m : Maybe A
        f : A -> Maybe B
        g : B -> Maybe C

left-identity  : return a >>= f ≡ f a
right-identity : m >>= return ≡ m
associativity  : m >>= (λ a -> f a >>= g) ≡ (m >>= f) >>= g

left-identity = reflexive

right-identity {m = Nothing} = reflexive
right-identity {m = Just a } = reflexive

associativity {m = Nothing} = reflexive
associativity {m = Just a } = reflexive
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

A useful twist of the left-identity law for equational reasoning (see
Huffman encoding). This looks quite innocent but it turned out to be
one of the most useful proof combinators.

<<invoke>>
\begin{code}
invoke : m ≡ return a -> m >>= f ≡ f a
invoke reflexive = reflexive

invoke-left-identity = invoke
\end{code}

map-congruent : ∀ {f₁ f₂ : A -> B} {m} -> (∀ a -> f₁ a ≡ f₂ a) -> 𝕄.map f₁ m ≡ 𝕄.map f₂ m
map-congruent {m = Nothing} feq = reflexive
map-congruent {m = Just x} feq = congruent Just (feq x)

\begin{code}
invoke-right-identity : ∀ {f : A -> Maybe A} (m : Maybe A) -> (∀ (a : A) -> f a ≡ return a) -> m >>= f ≡ m
invoke-right-identity (Nothing) f≡return = reflexive
invoke-right-identity (Just x)  f≡return = f≡return x

>>=-congruent : ∀ {m : Maybe A} {k₁ k₂ : A -> Maybe B} -> (∀ a -> k₁ a ≡ k₂ a) -> m >>= k₁ ≡ m >>= k₂
>>=-congruent {m = Nothing} feq = reflexive
>>=-congruent {m = Just x} feq = feq x

map->>= : ∀ {m : Maybe A} {f : A -> B} {k : B -> Maybe C} -> m >>= (λ a -> k (f a)) ≡ map f m >>= k
map->>= {m = Nothing} = reflexive
map->>= {m = Just x} = reflexive

>>=-map : ∀ {m : Maybe A} {f : B -> C} {k : A -> Maybe B} -> map f (m >>= k) ≡ m >>= λ a -> map f (k a)
>>=-map {m = Nothing} = reflexive
>>=-map {m = Just x} = reflexive
\end{code}

>>=-congruent₁ : ∀ {m₁ m₂ : Maybe A} -> m₁ ≡ m₂ -> (k : A -> Maybe B) -> m₁ 𝕄.>>= k ≡ m₂ 𝕄.>>= k
>>=-congruent₁ reflexive k  = reflexive

\begin{code}
\end{code}

-------------------------------------------------------------------------------

Further reading:

⚫  https://agda.readthedocs.io

„“ “”
