Properties of the strict ordering(s).

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Natural.Order.Properties where

open import VFP.Natural.Properties hiding (transitive) public
open import VFP.Natural.Order public
open import VFP.Equality.Properties hiding (transitive) public
open import VFP.Logic.Properties
open import VFP.Decidable

private
  variable
    a a₁ a₂ b b₁ b₂ k m n p : ℕ
\end{code}

-------------------------------------------------------------------------------

Properties of the strict order |≺|.

<<irreflexive>>
\begin{code}
≺-irreflexive : ¬ n ≺ n
≺-irreflexive (succ≺succ n≺n) = ≺-irreflexive n≺n
\end{code}

<<transitive>>
\begin{code}
≺-transitive : k ≺ m → m ≺ n → k ≺ n
≺-transitive (zero≺succ)     (succ≺succ _  ) = zero≺succ
≺-transitive (succ≺succ k≺m) (succ≺succ m≺n) = succ≺succ (≺-transitive k≺m m≺n)
\end{code}

Uniqueness of proofs.

\begin{code}
≺-unique : ∀ {m n} -> (p : m ≺ n) -> (q : m ≺ n) -> p ≡ q
≺-unique (zero≺succ)   (zero≺succ)                        = reflexive
≺-unique (succ≺succ p) (succ≺succ q) rewrite ≺-unique p q = reflexive
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Private stuff for lecture slides and sessions.

\begin{code}
private
\end{code}

<<3-precedes-5>>
\begin{code}
  three-precedes-five : 3 ≺ 5
  three-precedes-five = succ≺succ (succ≺succ (succ≺succ zero≺succ))
\end{code}

<<not-5-precedes-3>>
\begin{code}
  ¬five-precedes-three : ¬ 5 ≺ 3
  ¬five-precedes-three (succ≺succ (succ≺succ (succ≺succ ())))
\end{code}

<<transitive-private>>
\begin{code}
  transitive : k ≺ m → m ≺ n → k ≺ n
  transitive (zero≺succ)      (succ≺succ _  )  = zero≺succ
  transitive (succ≺succ k≺m)  (succ≺succ m≺n)  = succ≺succ (transitive k≺m m≺n)
\end{code}

\begin{code}
¬n≺zero : ¬ n ≺ zero
¬n≺zero ()
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Exclusiveness/no confusion.

<<decide-helper>>
\begin{code}
not-equal : ∀ {m n} -> m ≺ n -> ¬ m ≡ n
not-equal (zero≺succ)     = λ ()
not-equal (succ≺succ m≺n) = λ where
  reflexive → not-equal m≺n reflexive

not-equal′ : ∀ {m n} -> m ≺ n -> ¬ m ≡ n
not-equal′ p reflexive = ≺-irreflexive p

not-greater : ∀ {m n} -> m ≺ n -> ¬ m ≻ n
not-greater a≺b b≺a = ≺-irreflexive (≺-transitive a≺b b≺a)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Simple variant for the lectures.

\begin{code}
module Private where
\end{code}

<<Possibly>>
\begin{code}
  data Possibly (Property : Set) : Set where
    no  :              Possibly Property
    yes : Property ->  Possibly Property
\end{code}

<<decide-Possibly>>
\begin{code}
  _≺?_ : ∀ m n -> Possibly (m ≺ n)
  m       ≺? zero    = no
  zero    ≺? succ n  = yes zero≺succ
  succ m  ≺? succ n with m ≺? n
  ... | no           = no
  ... | yes m≺n      = yes (succ≺succ m≺n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Decidability.

Inequality is decidable i.e. given two natural numbers we can either
construct a proof of m ≺ n or a proof of ¬ m ≺ n.

<<decide>>
\begin{code}
infix 6 _≺?_
_≺?_ : ∀ m n -> Decide (m ≺ n)
m ≺? n with compare m n
... | LT {ev = m≺n}       = yes m≺n
... | EQ {ev = reflexive} = no ≺-irreflexive
... | GT {ev = m≻n}       = no (λ m≺n → ≺-irreflexive (≺-transitive m≺n m≻n))

infix 6 _≻?_
_≻?_ : ∀ m n -> Decide (m ≻ n)
a ≻? b = b ≺? a
\end{code}

m ≺? n with compare m n
... | LT {ev = m≺n}       = yes m≺n
... | EQ {ev = m≡n} = no  (flip not-equal m≡n)
... | GT {ev = m≻n} = no  (flip not-greater m≻n)

\begin{code}
≺-then-≼ : ∀ {a b c} → a ≺ b → b ≼ c → a ≺ c
≺-then-≼ a≺b (inl reflexive) = a≺b
≺-then-≼ a≺b (inr b≺c)       = ≺-transitive a≺b b≺c

≼-then-≺ : ∀ {a b c} → a ≼ b → b ≺ c → a ≺ c
≼-then-≺ (inl reflexive) b≺c = b≺c
≼-then-≺ (inr a≺b)       b≺c = ≺-transitive a≺b b≺c

infix 6 _≼?_
_≼?_ : ∀ m n -> Decide (m ≼ n)
m ≼? n with compare m n
... | LT {ev = m≺n} = yes (inr m≺n)
... | EQ {ev = m≡n} = yes (inl m≡n)
... | GT {ev = m≻n} = no (λ m≼n → ≺-irreflexive (≺-then-≼ m≻n m≼n))

\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of the non-strict order ≼.

\begin{code}
≼-reflexive : n ≼ n
≼-reflexive = inl reflexive

≼-transitive : k ≼ m → m ≼ n → k ≼ n
≼-transitive (inl reflexive) (k≼n)           = k≼n
≼-transitive (inr k≺m)       (inl reflexive) = inr k≺m
≼-transitive (inr k≺m)       (inr m≺n)       = inr (≺-transitive k≺m m≺n)

≼-antisymmetric : m ≼ n → n ≼ m → m ≡ n
≼-antisymmetric (inl reflexive) n≼m             = reflexive
≼-antisymmetric (inr m≺n)       (inl reflexive) = reflexive
≼-antisymmetric (inr m≺n)       (inr n≺m)       = ex-falso-quodlibet (not-greater m≺n n≺m)
\end{code}

\begin{code}
≽-reflexive = ≼-reflexive

≽-transitive : ∀ {k m n} → k ≽ m → m ≽ n → k ≽ n
≽-transitive p q = ≼-transitive q p
\end{code}

|succ≼succ| shows that the successor function is monotone.

\begin{code}
zero≼n : ∀ n -> zero ≼ n
zero≼n (zero)   = inl reflexive
zero≼n (succ n) = inr zero≺succ
 
succ≼succ : m ≼ n -> succ m ≼ succ n
succ≼succ (inl reflexive) = inl reflexive
succ≼succ (inr y)         = inr (succ≺succ y)

succ-strict-order-reflecting : succ m ≺ succ n -> m ≺ n
succ-strict-order-reflecting (succ≺succ p) = p

succ-order-reflecting : succ m ≼ succ n -> m ≼ n
succ-order-reflecting (inl reflexive)     = inl reflexive
succ-order-reflecting (inr y) = inr (succ-strict-order-reflecting y)

succ-reflects-≼ = succ-order-reflecting

n≼succ-n : n ≼ succ n
n≼succ-n = inr n≺succ-n
\end{code}

begin{code}
zero-bottom : ∀ {n} -> zero ≼ n
zero-bottom {zero} = inl reflexive
zero-bottom {succ n} = inr zero≺succ
end{code}

Decidability.

\begin{code}
decide-≼ : ∀ m n -> Decide (m ≼ n)
decide-≼ m n with compare m n
... | LT {ev = m≺n} = yes (inr m≺n)
... | EQ {ev = m≡n} = yes (inl m≡n)
... | GT {ev = m≻n} = no λ where
  (inl m≡n) → not-equal m≻n (symmetric m≡n) -- OR: ≺-irreflexive m≻n
  (inr m≺n) → not-greater m≺n m≻n

-- ad-hoc definition from scratch
_≼!_ : ∀ m n -> Decide (m ≼ n)
zero   ≼! zero   = yes (inl reflexive)
zero   ≼! succ j = yes (inr zero≺succ)
succ i ≼! zero   = no λ where
                        (inl ())
                        (inr ())
succ i ≼! succ j with i ≼! j
... | no refute           = no λ where
                                 (inl reflexive)       → refute (inl reflexive)
                                 (inr (succ≺succ i≺j)) → refute (inr i≺j)
... | yes (inl reflexive) = yes (inl reflexive)
... | yes (inr i≺j)       = yes (inr (succ≺succ i≺j))
\end{code}

private
  infix 6 _≼?_
  _≼?_ : ∀ m n → m ≼ n ⊎ n ≼ m
  zero   ≼? zero   = inl (inl reflexive)
  zero   ≼? succ n = inl (inr zero≺succ)
  succ m ≼? zero   = inr (inr zero≺succ)
  succ m ≼? succ n with m ≼? n
  ... | inl (inl reflexive) = inl (inl reflexive)
  ... | inl (inr m≺n)       = inl (inr (succ≺succ m≺n))
  ... | inr (inl reflexive) = inl (inl reflexive)
  ... | inr (inr n≺m)       = inr (inr (succ≺succ n≺m))

\begin{code}
total : ∀ m n → m ≼ n ⊎ m ≽ n
total zero zero = inl (inl reflexive)
total zero (succ n) = inl (inr zero≺succ)
total (succ m) zero = inr (inr zero≺succ)
total (succ m) (succ n) with total m n
... | inl m≼n = inl (succ≼succ m≼n)
... | inr n≼m = inr (succ≼succ n≼m)
\end{code}

\begin{code}
infix 6 _≼-or-≻_ _≼?≻_ _≺-or-≽_ _≺?≽_
_≼-or-≻_ _≼?≻_ : ∀ m n -> m ≼ n ⊎ m ≻ n
m ≼-or-≻ n with compare m n
... | LT {ev = m≺n} = inl (inr m≺n)
... | EQ {ev = m≡n} = inl (inl m≡n)
... | GT {ev = m≻n} = inr m≻n

_≼?≻_ = _≼-or-≻_

_≺-or-≽_ : ∀ m n -> m ≺ n ⊎ m ≽ n
m ≺-or-≽ n with n ≼-or-≻ m
... | inl n≼m = inr n≼m
... | inr m≻n = inl m≻n

_≺?≽_ = _≺-or-≽_
\end{code}

\begin{code}
infix 6 _≼-or-≽_ _≼?≽_
_≼-or-≽_ _≼?≽_ : ∀ m n -> m ≼ n ⊎ m ≽ n
m ≼-or-≽ n with compare m n
... | LT {ev = m≺n} = inl (inr m≺n)
... | EQ {ev = m≡n} = inl (inl m≡n)
... | GT {ev = m≻n} = inr (inr m≻n)

_≼?≽_ = _≼-or-≽_
\end{code}

\begin{code}
infix 6 _⩽-or-≻_
_⩽-or-≻_ : ∀ m n -> m ⩽ n ⊎ m ≻ n
m ⩽-or-≻ n with compare m n
... | LT {ev = m≺n      } = inl (LT m≺n)
... | EQ {ev = reflexive} = inl EQ
... | GT {ev = m≻n      } = inr m≻n
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Relating strict and non-strict ordering (reverse implication of
not-equal and not-greater).
 
\begin{code}
not-leq ¬≼-is-≻ : ∀ {m n} -> ¬ m ≼ n -> m ≻ n -- n ≺ m
not-leq {zero}   {zero}   p = ex-falso-quodlibet (p (inl reflexive))
not-leq {zero}   {succ n} p = ex-falso-quodlibet (p (inr zero≺succ))
not-leq {succ m} {zero}   p = zero≺succ
not-leq {succ m} {succ n} p = succ≺succ (not-leq λ where
                                                   (inl reflexive) → p (inl reflexive)
                                                   (inr m≺n)       → p (inr (succ≺succ m≺n)))
¬≼-is-≻ = not-leq

not-less : ∀ {m n} -> ¬ m ≼ n -> n ≼ m  -- TODO: rename
not-less ¬m≼n = inr (not-leq ¬m≼n)

¬≺-is-≽ : ∀ {m n} -> ¬ m ≺ n -> m ≽ n
¬≺-is-≽ {zero} {zero} p = inl reflexive
¬≺-is-≽ {zero} {succ n} p = ex-falso-quodlibet (p zero≺succ)
¬≺-is-≽ {succ m} {zero} p = inr zero≺succ
¬≺-is-≽ {succ m} {succ n} p with ¬≺-is-≽ {m} {n} (λ x → p (succ≺succ x))
... | inl reflexive = inl reflexive
... | inr y = inr (succ≺succ y)

≼-is-¬≻ : ∀ {m n} -> m ≼ n -> ¬ m ≻ n
≼-is-¬≻ {succ m} {zero} (inl ()) m≻n
≼-is-¬≻ {succ m} {zero} (inr ()) m≻n
≼-is-¬≻ {succ m} {succ .m} (inl reflexive) (succ≺succ m≻n) = ≺-irreflexive m≻n
≼-is-¬≻ {succ m} {succ n} (inr (succ≺succ y)) (succ≺succ m≻n) = ≼-is-¬≻ (inr y) m≻n

≺-is-¬≽ : ∀ {m n} -> m ≺ n -> ¬ m ≽ n
≺-is-¬≽ p q = ≼-is-¬≻ q p
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Impossible combinations.

\begin{code}
impossible₁ : ∀ {A : Set} -> ∀ {n} → n ≺ n → A
impossible₁ n≺n = ex-falso-quodlibet (≺-irreflexive n≺n)

impossible : ∀ {A : Set} -> ∀ {m n} → m ≺ n → n ≺ m -> A
impossible p q = ex-falso-quodlibet (not-greater q p)
--  impossible a≺b b≺a = ex-falso-quodlibet (≺-irreflexive (≺-transitive a≺b b≺a))

impossible₂ : ∀ {A : Set} -> ∀ {m n} → m ≺ n → n ≼ m -> A
impossible₂ m≺n (inl reflexive) = ex-falso-quodlibet (≺-irreflexive m≺n)
impossible₂ m≺n (inr n≺m)       = ex-falso-quodlibet (not-greater m≺n n≺m)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Relating strict and non-strict ordering via arithmetic.

\begin{code}
succ-≼-is-≺ : succ m ≼ n -> m ≺ n
succ-≼-is-≺ (inl reflexive) = n≺succ-n
succ-≼-is-≺ (inr y)         = ≺-transitive n≺succ-n y

≺-is-succ-≼ : m ≺ n -> succ m ≼ n 
≺-is-succ-≼ (zero≺succ)     = succ≼succ (zero≼n _)
≺-is-succ-≼ (succ≺succ m≺n) = succ≼succ (≺-is-succ-≼ m≺n)

≼-is-≺-succ : m ≼ n -> m ≺ succ n
≼-is-≺-succ (inl reflexive) = n≺succ-n
≼-is-≺-succ (inr y)         = ≺-transitive y n≺succ-n

≺-succ-is-≼ : m ≺ succ n -> m ≼ n
≺-succ-is-≼ (zero≺succ)     = zero≼n _
≺-succ-is-≼ (succ≺succ m≺n) = ≺-is-succ-≼ m≺n
\end{code}

One- and two-point intervals.

\begin{code}
the-one : m ≼ n -> n ≺ 1 + m -> m ≡ n
the-one p q = ≼-antisymmetric p (≺-succ-is-≼ q)

one-of-two : n ≼ m -> m ≼ 1 + n -> m ≡ n ⊎ m ≡ 1 + n
one-of-two (inl x) (inl x₁) = inr x₁
one-of-two (inl x) (inr x₁) = inl (symmetric x)
one-of-two (inr x) (inl x₁) = inr x₁
one-of-two (inr x) (inr x₁) = ex-falso-quodlibet (≺-irreflexive (≺-then-≼ x (≺-succ-is-≼ x₁)))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Relating Boolean-valued and Set-valued comparisons.

\begin{code}
open import VFP.Boolean

<-to : ∀ {a b} -> a ≺ b -> So (a < b)
<-to (zero≺succ)   = yes
<-to (succ≺succ p) = <-to p

<-fro : ∀ {a b} -> So (a < b) -> a ≺ b
<-fro {zero}   {succ b} yes = zero≺succ
<-fro {succ a} {succ b} p   = succ≺succ (<-fro p)

≥-fro : ∀ {a b} -> So (a ≥ b) -> a ≽ b
≥-fro {zero} {zero} yes = inl reflexive
≥-fro {succ a} {zero} yes = inr zero≺succ
≥-fro {succ a} {succ b} p = succ≼succ (≥-fro p)
\end{code}

-------------------------------------------------------------------------------

Properties of arithmetic operations (monotonicity, distributivity, etc).

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of minimum ↓.

\begin{code}
↓-idempotent : ∀ n → n ↓ n ≡ n
↓-idempotent n with compare n n
… | LT {ev = ev} = ex-falso-quodlibet (≺-irreflexive ev)
… | EQ {ev = ev} = reflexive
… | GT {ev = ev} = ex-falso-quodlibet (≺-irreflexive ev)
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Monotonicity: addition is (strictly) monotone in both arguments.

\begin{code}
private
  help : ∀ n k → k ≺ succ (n + k)
  help zero k = n≺succ-n
  help (succ n) k = ≺-transitive n≺succ-n (succ≺succ (help n k))

≺-monotone-left-+ : ∀ {k m n} -> m ≺ n -> m + k ≺ n + k
≺-monotone-left-+ zero≺succ = help _ _
≺-monotone-left-+ (succ≺succ p) = succ≺succ (≺-monotone-left-+ p)

≺-monotone-right-+ : ∀ {k m n} -> m ≺ n -> k + m ≺ k + n
≺-monotone-right-+ {k = zero}   p = p
≺-monotone-right-+ {k = succ k} p = succ≺succ (≺-monotone-right-+ p)

≺-monotone-+ : ∀ {m₁ m₂ n₁ n₂} -> m₁ ≺ m₂ → n₁ ≺ n₂ → m₁ + n₁ ≺ m₂ + n₂
≺-monotone-+ m₁≺m₂ n₁≺n₂ = ≺-transitive (≺-monotone-left-+ m₁≺m₂) (≺-monotone-right-+ n₁≺n₂)

≼-monotone-left-+ : ∀ {k m n} -> m ≼ n -> m + k ≼ n + k
≼-monotone-left-+ (inl reflexive) = inl reflexive
≼-monotone-left-+ (inr y) = inr (≺-monotone-left-+ y)

≼-monotone-right-+ : ∀ {k m n} -> m ≼ n -> k + m ≼ k + n
≼-monotone-right-+ (inl reflexive) = ≼-reflexive
≼-monotone-right-+ (inr y) = inr (≺-monotone-right-+ y)

+-left-preserves-≼  = ≼-monotone-left-+
+-right-preserves-≼ = ≼-monotone-right-+

k+-preserves-≼ : ∀ k {m n} -> m ≼ n -> k + m ≼ k + n
k+-preserves-≼ k = ≼-monotone-right-+ {k = k}

+-preserves-≼ : ∀ {m₁ m₂ n₁ n₂} → m₁ ≼ m₂ → n₁ ≼ n₂ → m₁ + n₁ ≼ m₂ + n₂
+-preserves-≼ m₁≼m₂ n₁≼n₂ = ≼-transitive (≼-monotone-left-+ m₁≼m₂) (≼-monotone-right-+ n₁≼n₂)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Monotonicity: multiplication is (strictly) monotone in both arguments.

\begin{code}
≺-monotone-left-* : ∀ {k m n} -> 0 ≺ k -> m ≺ n -> m * k ≺ n * k
≺-monotone-left-* zero≺succ zero≺succ = zero≺succ
≺-monotone-left-* zero≺succ (succ≺succ q) = succ≺succ (≺-monotone-right-+ (≺-monotone-left-* zero≺succ q))
\end{code}

TODO

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
[a↓b]*2≼a+b : ∀ a b → (a ↓ b) *2 ≼ a + b
[a↓b]*2≼a+b a b with compare a b
... | LT {ev = ev}        = ≼-transitive (inl (n*2≡n+n a)) (≼-monotone-right-+ (inr ev))
... | EQ {ev = reflexive} = inl (n*2≡n+n a)
... | GT {ev = ev}        = ≼-transitive (inl (n*2≡n+n b)) (≼-monotone-left-+ (inr ev))

a+b≼[a↑b]*2 : ∀ a b → a + b ≼ (a ↑ b) *2
a+b≼[a↑b]*2 a b with compare a b
... | LT {ev = ev}        = ≼-transitive (≼-monotone-left-+ (inr ev)) (inl (symmetric (n*2≡n+n b)))
... | EQ {ev = reflexive} = inl (symmetric (n*2≡n+n b))
... | GT {ev = ev}        = ≼-transitive (≼-monotone-right-+ (inr ev)) (inl (symmetric (n*2≡n+n a)))
\end{code}

\begin{code}
pred-monotonic : ∀ {m n : ℕ} → m ≼ n → pred m ≼ pred n
pred-monotonic (inl reflexive)       = inl reflexive
pred-monotonic (inr zero≺succ)       = zero≼n _
pred-monotonic (inr (succ≺succ m≺n)) = inr m≺n

≼-monotonic-*2 : ∀ {a b} → a ≼ b → a *2 ≼ b *2
≼-monotonic-*2 {zero} {b} p = zero≼n (b *2)
≼-monotonic-*2 {succ a} {zero} (inl ())
≼-monotonic-*2 {succ a} {zero} (inr ())
≼-monotonic-*2 {succ a} {succ .a} (inl reflexive) = inl reflexive
≼-monotonic-*2 {succ a} {succ b} (inr (succ≺succ y)) = succ≼succ (succ≼succ (≼-monotonic-*2 (inr y)))

*2-preserves-≼ = ≼-monotonic-*2

≺-monotonic-*2 : ∀ {a b} → a ≺ b → a *2 ≺ b *2
≺-monotonic-*2 {zero} {succ b} p = zero≺succ
≺-monotonic-*2 {succ a} {succ b} (succ≺succ p) = succ≺succ (succ≺succ (≺-monotonic-*2 p))
*2-preserves-≺ = ≺-monotonic-*2

⌊/2⌋-monotonic : ∀ {m n : ℕ} → m ≼ n → ⌊ m /2⌋ ≼ ⌊ n /2⌋
⌊/2⌋-monotonic {zero} {n} p = zero≼n ⌊ n /2⌋
⌊/2⌋-monotonic {succ m} {zero} (inl ())
⌊/2⌋-monotonic {succ m} {zero} (inr ())
⌊/2⌋-monotonic {succ zero} {succ zero} (inl reflexive) = inl reflexive
⌊/2⌋-monotonic {succ zero} {succ zero} (inr (succ≺succ ()))

⌊/2⌋-monotonic {succ zero} {succ (succ n)} p = inr zero≺succ
⌊/2⌋-monotonic {succ (succ m)} {succ zero} (inr (succ≺succ ()))
⌊/2⌋-monotonic {succ (succ m)} {succ (succ .m)} (inl reflexive) = succ≼succ (⌊/2⌋-monotonic (inl reflexive))
⌊/2⌋-monotonic {succ (succ m)} {succ (succ n)} (inr (succ≺succ (succ≺succ y))) = succ≼succ (⌊/2⌋-monotonic (inr y))

⌊/2⌋-monotonic-≺-≼ : ∀ {m n : ℕ} → m ≺ n → ⌊ m /2⌋ ≼ ⌊ n /2⌋
⌊/2⌋-monotonic-≺-≼ {zero         } {succ zero    } _                           = inl reflexive
⌊/2⌋-monotonic-≺-≼ {zero         } {succ (succ n)} _                           = inr zero≺succ
⌊/2⌋-monotonic-≺-≼ {succ zero    } {succ (succ n)} (succ≺succ zero≺succ)       = inr zero≺succ
⌊/2⌋-monotonic-≺-≼ {succ (succ m)} {succ (succ n)} (succ≺succ (succ≺succ m≺n)) = succ≼succ (⌊/2⌋-monotonic-≺-≼ {m} {n} m≺n)

⌊/2⌋-monotonic-*2≺ : ∀ {m n : ℕ} → m ≺ n *2 → ⌊ m /2⌋ ≺ n
⌊/2⌋-monotonic-*2≺ {zero         } {succ n} m≺*2n                         = zero≺succ
⌊/2⌋-monotonic-*2≺ {succ zero    } {succ n} m≺*2n                         = zero≺succ
⌊/2⌋-monotonic-*2≺ {succ (succ m)} {succ n} (succ≺succ (succ≺succ m≺n*2)) = succ≺succ (⌊/2⌋-monotonic-*2≺ m≺n*2)
\end{code}

\begin{code}
private
  base₀ : ∀ n → 0 ≺ 2^ n
  base₀ zero = zero≺succ
  base₀ (succ n) = *2-preserves-≺ (base₀ n)

  base₁ : ∀ n → 1 ≼ 2^ n
  base₁ n = ≺-succ-is-≼ (succ≺succ (base₀ n))

≼-monotonic-2^ : ∀ {a b} → a ≺ b → 2^ a ≺ 2^ b
≼-monotonic-2^ {zero} {succ b} zero≺succ = ≺-then-≼ (succ≺succ zero≺succ) (≼-monotonic-*2 (base₁ b))
≼-monotonic-2^ {succ a} {succ b} (succ≺succ p) = ≺-monotonic-*2 (≼-monotonic-2^ p)

2^-preserves-≺ = ≼-monotonic-2^
\end{code}

Feijen proof of base case (proof format cannot be used because of
cyclic module dependencies):

    proof
      1
    ≺⟨ succ≺succ zero≺succ ⟩
      2
    ≼⟨ ≼-monotonic-*2 (base₁ b) ⟩
     (2^ b) *2
    ∎ where open import VFP.Natural.Order.Reasoning

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of monus ∸.

\begin{code}
m∸n≼m : ∀ {m n} -> m ∸ n ≼ m
m∸n≼m {zero}   {zero}   = inl reflexive
m∸n≼m {zero}   {succ n} = m∸n≼m {zero} {n}
m∸n≼m {succ m} {zero}   = inl reflexive
m∸n≼m {succ m} {succ n} = (≼-transitive (m∸n≼m) (inr n≺succ-n))
\end{code}

TODO: polish proofs.

\begin{code}
∸-preserves-≼ : ∀ a b n -> a ≼ b -> a ∸ n ≼ b ∸ n
∸-preserves-≼ zero zero zero p = p
∸-preserves-≼ zero zero (succ n) p = p
∸-preserves-≼ zero (succ b) zero p = p
∸-preserves-≼ zero (succ b) (succ n) p = zero≼n (b ∸ n)
∸-preserves-≼ (succ a) zero zero p = p
∸-preserves-≼ (succ a) zero (succ n) (inl ())
∸-preserves-≼ (succ a) zero (succ n) (inr ())
∸-preserves-≼ (succ a) (succ b) zero p = p
∸-preserves-≼ (succ a) (succ b) (succ n) p = ∸-preserves-≼ a b n (succ-order-reflecting p)
\end{code}

TODO: "preserves" corresponds to "monotone"; "flips" (?) corresponds
to antitone".

\begin{code}
∸-flips-≼ : ∀ a b n -> b ≼ a -> n ∸ a ≼ n ∸ b
∸-flips-≼ zero zero zero p = p
∸-flips-≼ zero zero (succ n) p = inl reflexive
∸-flips-≼ zero (succ b) zero p = inl reflexive
∸-flips-≼ zero (succ b) (succ n) (inl ())
∸-flips-≼ zero (succ b) (succ n) (inr ())
∸-flips-≼ (succ a) zero zero p = inl reflexive
∸-flips-≼ (succ a) zero (succ n) p = ≼-transitive m∸n≼m (inr n≺succ-n)
∸-flips-≼ (succ a) (succ b) zero p = inl reflexive
∸-flips-≼ (succ a) (succ b) (succ n) p = ∸-flips-≼ a b n (succ-order-reflecting p)
\end{code}

Interaction with strict ordering.

\begin{code}
∸-positive : ∀ {m n} -> m ≺ n -> zero ≺ n ∸ m
∸-positive zero≺succ = zero≺succ
∸-positive (succ≺succ p) = ∸-positive p

∸-pseudo-monotone : ∀ {a b n} -> n ≺ b -> a ≺ b -> a ∸ n ≺ b ∸ n
∸-pseudo-monotone (zero≺succ) (zero≺succ) = zero≺succ
∸-pseudo-monotone (zero≺succ) (succ≺succ q) = succ≺succ q
∸-pseudo-monotone (succ≺succ p) (zero≺succ) = ∸-positive p
∸-pseudo-monotone (succ≺succ p) (succ≺succ q) = ∸-pseudo-monotone p q

∸-pseudo-antitone : ∀ {a b n} -> b ≺ n -> b ≺ a -> n ∸ a ≺ n ∸ b
∸-pseudo-antitone zero≺succ zero≺succ = ≼-is-≺-succ m∸n≼m
∸-pseudo-antitone (succ≺succ p) (succ≺succ q) = ∸-pseudo-antitone p q
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
↓-monotone : ∀ {a₁ a₂} → a₁ ≼ a₂ → ∀ {b₁ b₂} → b₁ ≼ b₂ → a₁ ↓ b₁ ≼ a₂ ↓ b₂
↓-monotone {a₁ = a₁} {a₂ = a₂} p {b₁ = b₁} {b₂ = b₂} q with compare a₁ b₁ | compare a₂ b₂
… | LT {ev = ev₁} | LT {ev = ev₂} = p
… | LT {ev = ev₁} | EQ {ev = reflexive} = p
… | LT {ev = ev₁} | GT = ≼-transitive (inr ev₁) q
… | EQ {ev = reflexive} | LT = p
… | EQ {ev = reflexive} | EQ = p
… | EQ {ev = reflexive} | GT = q
… | GT {ev = ev₁} | LT {ev = ev₂} = ≼-transitive (inr ev₁) p
… | GT {ev = ev₁} | EQ {ev = reflexive} = q
… | GT {ev = ev₁} | GT {ev = ev₂} = q

↑-monotone : ∀ {a₁ a₂} → a₁ ≼ a₂ → ∀ {b₁ b₂} → b₁ ≼ b₂ → a₁ ↑ b₁ ≼ a₂ ↑ b₂
↑-monotone {a₁ = a₁} {a₂ = a₂} p {b₁ = b₁} {b₂ = b₂} q with compare a₁ b₁ | compare a₂ b₂
… | LT {ev = ev₁} | LT {ev = ev₂} = q
… | LT {ev = ev₁} | EQ {ev = reflexive} = q
… | LT | GT {ev = ev₁} = ≼-transitive q (inr ev₁)
… | EQ {ev = reflexive} | LT = q
… | EQ {ev = reflexive} | EQ = q
… | EQ {ev = reflexive} | GT = p
… | GT {ev = ev₁} | LT {ev = ev₂} = ≼-transitive p (inr ev₂)
… | GT {ev = ev₁} | EQ {ev = reflexive} = p
… | GT {ev = ev₁} | GT {ev = ev₂} = p
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Distributivity aka homomorphisms.

succ-distributes-↓ : ∀ {a b} → succ a ↓ succ b ≡ succ (a ↓ b)
succ-distributes-↓ {a} {b} with compare a b
… | LT = reflexive
… | EQ {ev = reflexive} = reflexive
… | GT = reflexive

\begin{code}
succ-preserves-↓ : ∀ {a b} → succ (a ↓ b) ≡ succ a ↓ succ b
succ-preserves-↓ {a} {b} with compare a b
... | LT                  = reflexive
... | EQ {ev = reflexive} = reflexive
... | GT                  = reflexive

succ-preserves-↑ : ∀ {a b} → succ (a ↑ b) ≡ succ a ↑ succ b
succ-preserves-↑ {a} {b} with compare a b
... | LT                  = reflexive
... | EQ {ev = reflexive} = reflexive
... | GT                  = reflexive
\end{code}

\begin{code}
+-distributes-↓ : ∀ n {a b} → (n + a) ↓ (n + b) ≡ n + (a ↓ b)
+-distributes-↓ zero = reflexive
+-distributes-↓ (succ n) = ≡-transitive (symmetric succ-preserves-↓) (congruent succ (+-distributes-↓ n))

*2-distributes-↓ : ∀ a b → (a *2) ↓ (b *2) ≡ (a ↓ b) *2
*2-distributes-↓ zero zero = reflexive
*2-distributes-↓ zero (succ b) = reflexive
*2-distributes-↓ (succ a) zero = reflexive
*2-distributes-↓ (succ a) (succ b) = ≡-transitive (symmetric succ-preserves-↓) (≡-transitive (congruent succ (symmetric succ-preserves-↓)) (≡-transitive (congruent (λ hole → 2 + hole) (*2-distributes-↓ a b)) (congruent _*2 succ-preserves-↓)))
\end{code}

\begin{code}
2^-preserves-↓ : ∀ a b → 2^ (a ↓ b) ≡ (2^ a) ↓ (2^ b)
2^-preserves-↓ a b with compare a b
2^-preserves-↓ a b | LT {ev = ev} with compare (2^ a) (2^ b)
... | LT            = reflexive
... | EQ            = reflexive
... | GT {ev = ev′} = impossible (2^-preserves-≺ ev) ev′
2^-preserves-↓ a b | EQ {ev = reflexive} = symmetric (↓-idempotent (2^ a))
2^-preserves-↓ a b | GT {ev = ev} with compare (2^ a) (2^ b)
... | LT {ev = ev′} = impossible ev′ (2^-preserves-≺ ev)
... | EQ {ev = ev′} = symmetric ev′
... | GT            = reflexive

2^-distributes-↓ = 2^-preserves-↓
\end{code}

\begin{code}
2^-preserves-↑ : ∀ a b → 2^ (a ↑ b) ≡ (2^ a) ↑ (2^ b)
2^-preserves-↑ a b with compare a b | compare (2^ a) (2^ b)
... | LT | LT = reflexive
... | LT | EQ = reflexive
... | LT {ev = ev₁} | GT {ev = ev₂} = impossible (2^-preserves-≺ ev₁) ev₂
... | EQ | LT = reflexive
... | EQ | EQ = reflexive
... | EQ {ev = reflexive} | GT = reflexive
... | GT {ev = ev₁} | LT {ev = ev₂} = impossible (2^-preserves-≺ ev₁) ev₂
... | GT {ev = ev₁} | EQ {ev = ev₂} = ev₂
... | GT | GT = reflexive
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Properties of natural division by 2.
TODO: clean up.

\begin{code}
⌊n/2⌋≺succ-n : ∀ n → ⌊ n /2⌋ ≺ succ n
⌈n/2⌉≺succ-n : ∀ n → ⌈ n /2⌉ ≺ succ n

⌊n/2⌋≺succ-n (zero)   = zero≺succ
⌊n/2⌋≺succ-n (succ n) = ≺-transitive (⌈n/2⌉≺succ-n n) n≺succ-n

--⌊n/2⌋≺succ-n zero = zero≺succ
--⌊n/2⌋≺succ-n (succ zero) = zero≺succ
--⌊n/2⌋≺succ-n (succ (succ n)) = succ≺succ (≺-transitive (⌊n/2⌋≺succ-n n) n≺succ-n)

⌈n/2⌉≺succ-n (zero)   = zero≺succ
⌈n/2⌉≺succ-n (succ n) = succ≺succ (⌊n/2⌋≺succ-n n)

--⌊succ-n/2⌋≺succ-n : ∀ n → ⌊ succ n /2⌋ ≺ succ n
--⌊succ-n/2⌋≺succ-n zero = zero≺succ
--⌊succ-n/2⌋≺succ-n (succ n) = succ≺succ (⌊n/2⌋≺succ-n n)

⌊n/2⌋≼n : ∀ n → ⌊ n /2⌋ ≼ n
⌈n/2⌉≼n : ∀ n → ⌈ n /2⌉ ≼ n

⌊n/2⌋≼n n = ≺-succ-is-≼ (⌊n/2⌋≺succ-n n)
⌈n/2⌉≼n n = ≺-succ-is-≼ (⌈n/2⌉≺succ-n n)

--⌈n/2⌉≼n = ≺-succ-is-≼ ⌊n/2⌋≺succ-n

--⌈n/2⌉≼n : ∀ n → ⌈ n /2⌉ ≼ n
--⌈n/2⌉≼n zero            = inl reflexive
--⌈n/2⌉≼n (succ zero)     = inl reflexive
--⌈n/2⌉≼n (succ (succ n)) = succ≼succ (≼-transitive (⌈n/2⌉≼n n) (inr n≺succ-n))
\end{code}

⌊-/2⌋-decreasing : ∀ n -> ⌊ n /2⌋ ≼ n
⌊-/2⌋-decreasing (zero) = zero≼n
⌊-/2⌋-decreasing (succ zero) = zero≼n
⌊-/2⌋-decreasing (succ (succ n)) = succ≼succ (≼-transitive (⌊-/2⌋-decreasing n) (inr n≺succ-n))

⌈-/2⌉-decreasing : ∀ n -> ⌈ n /2⌉ ≼ n  -- ⌊ succ n /2⌋ ≡ ⌈ n /2⌉
⌈-/2⌉-decreasing zero = zero≼n
⌈-/2⌉-decreasing (succ n) = succ≼succ (⌊-/2⌋-decreasing n)

Goal: 1 ≺ n → ⌊ n /2⌋ ≻ 0
Goal: 1 ≺ n → ⌊ n /2⌋ ≺ n

Useful for binary subdivision scheme: |n ↦ ⌊ n /2⌋ + ⌈ n /2⌉|.

\begin{code}
⌊n/2⌋≺n : ∀ {n} → n ≻ 1 → ⌊ n /2⌋ ≺ n
⌊n/2⌋≺n {succ (succ zero)}     (succ≺succ p) = succ≺succ p
⌊n/2⌋≺n {succ (succ (succ n))} (succ≺succ p) = succ≺succ (⌊n/2⌋≺succ-n (succ n))

⌈n/2⌉≺n : ∀ {n} → n ≻ 1 → ⌈ n /2⌉ ≺ n
⌈n/2⌉≺n {succ (succ zero)} (succ≺succ p) = succ≺succ p
⌈n/2⌉≺n {succ (succ (succ n))} (succ≺succ p) = succ≺succ (succ≺succ (⌊n/2⌋≺succ-n n))

⌊n/2⌋≻0 : ∀ {n} → 1 ≺ n → ⌊ n /2⌋ ≻ 0
⌊n/2⌋≻0 {succ (succ n)} (succ≺succ p) = zero≺succ

⌈n/2⌉≻0 : ∀ {n} → 1 ≺ n → ⌈ n /2⌉ ≻ 0
⌈n/2⌉≻0 {succ (succ n)} (succ≺succ p) = ⌊n/2⌋≻0 (succ≺succ zero≺succ)
\end{code}


-------------------------------------------------------------------------------

Properties of |≺₁|.

\begin{code}
≺₁-transitive : k ≺₁ m → m ≺₁ n → k ≺₁ n
≺₁-transitive p (one)    = succ p
≺₁-transitive p (succ q) = succ (≺₁-transitive p q)
\end{code}

<<zero-bottom-1>>
\begin{code}
zero-bottom₁ : ∀ {n} -> zero ≺₁ succ n
zero-bottom₁ {zero}   = one
zero-bottom₁ {succ n} = succ (zero-bottom₁ {n})
\end{code}

<<succ-monotone-1>>
\begin{code}
succ-monotone₁ : ∀ {m n} -> m ≺₁ n → succ m ≺₁ succ n   -- difference is unchanged
succ-monotone₁ (one)       = one
succ-monotone₁ (succ m≺₁n) = succ (succ-monotone₁ m≺₁n)
\end{code}

\begin{code}
-- order embedding₁: m ≺₁ n  <->  succ m ≺₁ succ n
embedding₁ : ∀ {m n} -> succ m ≺₁ succ n → m ≺₁ n   -- ditto
embedding₁ {m} {succ m} (one)    = one
embedding₁ {m} {succ n} (succ p) = succ (embedding₁ p)
\end{code}

This version is equivalent to |≺|.

<<to-1>>
\begin{code}
to-≺₁ : ∀ {m n} -> m ≺ n -> m ≺₁ n
to-≺₁ (zero≺succ)     = zero-bottom₁
to-≺₁ (succ≺succ m≺n) = succ-monotone₁ (to-≺₁ m≺n)
\end{code}

<<fro-1>>
\begin{code}
fro-≺₁ : ∀ {m n} -> m ≺₁ n -> m ≺ n
fro-≺₁ (one)    = n≺succ-n
fro-≺₁ (succ p) = ≺-transitive (fro-≺₁ p) n≺succ-n
\end{code}

\begin{code}
not-equal₁ : ∀ m n -> m ≺₁ n -> ¬ m ≡ n
not-equal₁ zero (succ m) (one) = λ ()
not-equal₁ zero (succ n) (succ p) = λ ()
not-equal₁ (succ m) (succ n) p = λ where
  reflexive → not-equal₁ m n (embedding₁ p) reflexive

≺₁-irreflexive : ∀ n -> ¬ n ≺₁ n
≺₁-irreflexive n p = not-equal₁ n n p reflexive
\end{code}

--------------------------------------------------------------------------------

Properties of |≼₂|.

\begin{code}
≼₂-transitive : k ≼₂ m → m ≼₂ n → k ≼₂ n
≼₂-transitive (zero)   q = q
≼₂-transitive (succ p) q = succ (≼₂-transitive p q)

n≼₂succ-n : ∀ {n} -> n ≼₂ succ n
n≼₂succ-n = succ zero

zero≼₂n : ∀ {n} -> zero ≼₂ n
zero≼₂n {zero}   = zero
zero≼₂n {succ n} = ≼₂-transitive (zero≼₂n {n}) (succ zero)

succ-monotone₂ : ∀ {m n} -> m ≼₂ n → succ m ≼₂ succ n   -- difference is unchanged
succ-monotone₂ (zero)   = zero
succ-monotone₂ (succ p) = succ (succ-monotone₂ p)
\end{code}

\begin{code}
to-≼₂ : ∀ {m n} -> m ≼ n -> m ≼₂ n
to-≼₂ (inl reflexive) = zero
to-≼₂ (inr zero≺succ) = zero≼₂n
to-≼₂ (inr (succ≺succ y)) = succ-monotone₂ (to-≼₂ (inr y))

fro-≼₂ : ∀ {m n} -> m ≼₂ n -> m ≼ n
fro-≼₂ (zero)   = ≼-reflexive
fro-≼₂ (succ p) = ≼-transitive (inr n≺succ-n) (fro-≼₂ p)

to-≺ : ∀ {m n} → succ m ≼₂ n → m ≺ n
to-≺ p with fro-≼₂ p
... | inl reflexive = n≺succ-n
... | inr y = ≺-transitive n≺succ-n y
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₃|.

\begin{code}
zero-bottom₃ : ∀ {n} -> zero ≺₃ succ n
zero-bottom₃ {zero}   = one
zero-bottom₃ {succ n} = ≺₃-transitive zero-bottom₃ one

succ-monotone₃ : ∀ {m n} -> m ≺₃ n → succ m ≺₃ succ n   -- difference is unchanged
succ-monotone₃ (one)               = one
succ-monotone₃ (≺₃-transitive p q) = ≺₃-transitive (succ-monotone₃ p) (succ-monotone₃ q)
\end{code}

This version is equivalent to |≺|.

\begin{code}
to-≺₃ : ∀ {m n} -> m ≺ n -> m ≺₃ n
to-≺₃ (zero≺succ)     = zero-bottom₃
to-≺₃ (succ≺succ m≺n) = succ-monotone₃ (to-≺₃ m≺n)

fro-≺₃ : ∀ {m n} -> m ≺₃ n -> m ≺ n
fro-≺₃ (one)                = n≺succ-n
fro-≺₃ (≺₃-transitive p p₁) = ≺-transitive (fro-≺₃ p) (fro-≺₃ p₁)
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₄|, the variant that relates ordering to addition
using a Σ-type.

\begin{code}
open import VFP.Natural.Properties
open import VFP.Equality.Reasoning

≺₄-transitive : ∀ {k m n} -> k ≺₄ m → m ≺₄ n → k ≺₄ n
≺₄-transitive {k = k} (Δ₁ , reflexive) (Δ₂ , reflexive) = succ (Δ₂ + Δ₁) , congruent succ (
    succ ((Δ₂ + Δ₁) + k)
  ≡⟨ +-associative (succ Δ₂) ⟩ 
    succ (Δ₂ + (Δ₁ + k))
  ≡⟨ symmetric (shift Δ₂ (Δ₁ + k)) ⟩ 
    Δ₂ + succ (Δ₁ + k)
  ∎ )
\end{code}

This version is equivalent to |≺₁|.

<<to-fro-4>>
\begin{code}
to-≺₄ : ∀ {m n} -> m ≺₁ n -> m ≺₄ n
to-≺₄ (one)    = 0 , reflexive
to-≺₄ (succ p) with to-≺₄ p
... | Δ , m≺₁n = succ Δ , congruent succ m≺₁n

fro-≺₄ : ∀ {m n} -> m ≺₄ n -> m ≺₁ n
fro-≺₄ (zero   , reflexive) = one
fro-≺₄ (succ Δ , reflexive) = succ (fro-≺₄ (Δ , reflexive))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
private
  lemma : ∀ (Δ₁ Δ₂ k : ℕ) -> succ ((Δ₂ + Δ₁) + k) ≡ Δ₂ + succ (Δ₁ + k)
  lemma Δ₁ Δ₂ k =
      succ ((Δ₂ + Δ₁) + k)
    ≡⟨ +-associative (succ Δ₂) ⟩ 
      succ (Δ₂ + (Δ₁ + k))
    ≡⟨ symmetric (shift Δ₂ (Δ₁ + k)) ⟩ 
      Δ₂ + succ (Δ₁ + k)
    ∎

≺₅-transitive : ∀ {k m n} -> k ≺₅ m → m ≺₅ n → k ≺₅ n
≺₅-transitive {k = k} (diff Δ₁) (diff Δ₂) =
  substitutive (λ hole → k ≺₅ succ hole) (lemma Δ₁ Δ₂ k) (diff (succ (Δ₂ + Δ₁)))
\end{code}

\begin{code}
to-≺₅ : ∀ {m n} -> m ≺₁ n -> m ≺₅ n
to-≺₅ (one)    = diff zero
to-≺₅ (succ p) with to-≺₅ p
... | diff Δ = diff (succ Δ)

fro-≺₅ : ∀ {m n} -> m ≺₅ n -> m ≺₁ n
fro-≺₅ (diff zero)     = one
fro-≺₅ (diff (succ Δ)) = succ (fro-≺₅ (diff Δ))
\end{code}

-------------------------------------------------------------------------------

Properties of |≺₆|.

\begin{code}
≺₆-transitive : ∀ {k m n} -> k ≺₆ m → m ≺₆ n → k ≺₆ n
≺₆-transitive {k = k} (Δ₁ , reflexive) (Δ₂ , reflexive) = succ (Δ₁ + Δ₂) , (
    succ (k + succ (Δ₁ + Δ₂))
  ≡⟨ congruent succ (shift k (Δ₁ + Δ₂)) ⟩ 
    succ (succ (k + (Δ₁ + Δ₂)))
  ≡⟨ congruent (λ hole → succ (succ hole)) (symmetric (+-associative k)) ⟩ 
    succ (succ ((k + Δ₁) + Δ₂))
  ∎ )
\end{code}

TODO: equivalence to |≺₁|.

-------------------------------------------------------------------------------

TODO: polish.

Connecting lemma: linking the order-theoretic and the
lattice-theoretic definitions.

\begin{code}
connecting-lemma-⊓ : a ≼ b → a ⊓ b ≡ a
connecting-lemma-⊓ (inl reflexive) = ⊓-idempotent _
connecting-lemma-⊓ (inr zero≺succ) = reflexive
connecting-lemma-⊓ (inr (succ≺succ x)) = congruent succ (connecting-lemma-⊓ (inr x))

connecting-lemma-⊓ᴿ : a ⊓ b ≡ a → a ≼ b
connecting-lemma-⊓ᴿ {a = zero} {b = zero} reflexive = inl reflexive
connecting-lemma-⊓ᴿ {a = zero} {b = succ b} eq = zero≼n (succ b)
connecting-lemma-⊓ᴿ {a = succ a} {b = succ b} eq = succ≼succ (connecting-lemma-⊓ᴿ (succ-injective eq))

connecting-lemma-⊓′ : a ≽ b → b ≡ a ⊓ b
connecting-lemma-⊓′ a≽b = ≡-transitive (symmetric (connecting-lemma-⊓ a≽b)) ((⊓-commutative _ _))
\end{code}

\begin{code}
connecting-lemma-⊔ : {a b : ℕ} → a ≼ b → a ⊔ b ≡ b
connecting-lemma-⊔ {zero} {b} p = reflexive
connecting-lemma-⊔ {succ a} {zero} (inl ())
connecting-lemma-⊔ {succ a} {zero} (inr ())
connecting-lemma-⊔ {succ a} {succ b} a≼b = congruent succ (connecting-lemma-⊔ (succ-order-reflecting a≼b))

connecting-lemma-⊔′ : {a b : ℕ} → a ≼ b → b ≡ b ⊔ a
connecting-lemma-⊔′ {a} {b} p = ≡-symmetric (≡-transitive (⊔-commutative b a) (connecting-lemma-⊔ p))

connecting-lemma-⊔ᴿ : {a b : ℕ} → a ⊔ b ≡ b → a ≼ b
connecting-lemma-⊔ᴿ {zero} {b} eq = zero≼n b
connecting-lemma-⊔ᴿ {succ a} {succ b} eq = succ≼succ (connecting-lemma-⊔ᴿ (succ-injective eq))
\end{code}

Other half of connecting lemma.

\begin{code}
meetl : a ⊓ b ≼ a
meetl {a = zero} {b = zero} = ≼-reflexive
meetl {a = zero} {b = succ b} = ≼-reflexive
meetl {a = succ a} {b = zero} = inr zero≺succ
meetl {a = succ a} {b = succ b} = succ≼succ meetl

meetr : a ⊓ b ≼ b
meetr {a = zero} {b = zero} = ≼-reflexive
meetr {a = zero} {b = succ b} = inr zero≺succ
meetr {a = succ a} {b = zero} = ≼-reflexive
meetr {a = succ a} {b = succ b} = succ≼succ meetr
\end{code}

\begin{code}
joinl : a ≼ a ⊔ b
joinl {a = zero} {b = zero} = ≼-reflexive
joinl {a = zero} {b = succ b} = inr zero≺succ
joinl {a = succ a} {b = zero} = ≼-reflexive
joinl {a = succ a} {b = succ b} = succ≼succ joinl

joinr : b ≼ a ⊔ b
joinr {b = zero} {a = zero} = ≼-reflexive
joinr {b = succ b} {a = zero} = ≼-reflexive
joinr {b = zero} {a = succ a} = inr zero≺succ
joinr {b = succ b} {a = succ a} = succ≼succ joinr
\end{code}

\begin{code}
meetl-preserves-≼ : a₁ ≼ a₂ → a₁ ⊓ b ≼ a₂ ⊓ b 
meetl-preserves-≼ {a₁ = zero} {a₂ = a₂} {b = b} p = zero≼n (a₂ ⊓ b)
meetl-preserves-≼ {a₁ = succ a₁} {a₂ = zero} {b = zero} p = ≼-reflexive
meetl-preserves-≼ {a₁ = succ a₁} {a₂ = zero} {b = succ b} (inl ())
meetl-preserves-≼ {a₁ = succ a₁} {a₂ = zero} {b = succ b} (inr ())
meetl-preserves-≼ {a₁ = succ a₁} {a₂ = succ a₂} {b = zero} p = ≼-reflexive
meetl-preserves-≼ {a₁ = succ a₁} {a₂ = succ a₂} {b = succ b} p = succ≼succ (meetl-preserves-≼ (succ-order-reflecting p))

meetr-preserves-≼ : b₁ ≼ b₂ → a ⊓ b₁ ≼ a ⊓ b₂ 
meetr-preserves-≼ {b₁ = b₁} {b₂ = b₂} {a = zero} p = ≼-reflexive
meetr-preserves-≼ {b₁ = zero} {b₂ = zero} {a = succ a} p = ≼-reflexive
meetr-preserves-≼ {b₁ = zero} {b₂ = succ b₂} {a = succ a} p = zero≼n (succ (a ⊓ b₂))
meetr-preserves-≼ {b₁ = succ b₁} {b₂ = zero} {a = succ a} (inl ())
meetr-preserves-≼ {b₁ = succ b₁} {b₂ = zero} {a = succ a} (inr ())
meetr-preserves-≼ {b₁ = succ b₁} {b₂ = succ b₂} {a = succ a} p = succ≼succ (meetr-preserves-≼ (succ-order-reflecting p))

meet-preserves-≼ : a₁ ≼ a₂ → b₁ ≼ b₂ → a₁ ⊓ b₁ ≼ a₂ ⊓ b₂
meet-preserves-≼ p q = ≼-transitive (meetl-preserves-≼ p) (meetr-preserves-≼ q)
\end{code}

\begin{code}
⊓-greatest-lower-bound : {a b c : ℕ} → c ≼ a → c ≼ b → c ≼ a ⊓ b
⊓-greatest-lower-bound {c = c} c≼a c≼b = ≼-transitive (inl (≡-symmetric (⊓-idempotent c))) (meet-preserves-≼ c≼a c≼b)
\end{code}

\begin{code}
joinl-preserves-≼ : {a₁ a₂ b : ℕ} → a₁ ≼ a₂ → a₁ ⊔ b ≼ a₂ ⊔ b
joinl-preserves-≼ {zero} {a₂} {b} a₁≼a₂ = joinr
joinl-preserves-≼ {succ a₁} {zero} {b} (inl ())
joinl-preserves-≼ {succ a₁} {zero} {b} (inr ())
joinl-preserves-≼ {succ a₁} {succ a₂} {zero} a₁≼a₂ = a₁≼a₂
joinl-preserves-≼ {succ a₁} {succ a₂} {succ b} a₁≼a₂ = succ≼succ (joinl-preserves-≼ (succ-order-reflecting a₁≼a₂))

joinr-preserves-≼ : {a b₁ b₂ : ℕ} → b₁ ≼ b₂ → a ⊔ b₁ ≼ a ⊔ b₂
joinr-preserves-≼ {a} {b₁} {b₂} b₁≼b₂ rewrite ⊔-commutative a b₁ rewrite ⊔-commutative a b₂ = joinl-preserves-≼ b₁≼b₂

join-preserves-≼ : {a₁ a₂ b₁ b₂ : ℕ} → a₁ ≼ a₂ → b₁ ≼ b₂ → a₁ ⊔ b₁ ≼ a₂ ⊔ b₂
join-preserves-≼ a₁≼a₂ b₁≼b₂ = ≼-transitive (joinl-preserves-≼ a₁≼a₂) (joinr-preserves-≼ b₁≼b₂)
\end{code}

\begin{code}
⊔-least-upper-bound : {a b c : ℕ} → a ≼ c → b ≼ c → a ⊔ b ≼ c
⊔-least-upper-bound {c = c} a≼c b≼c = ≼-transitive (join-preserves-≼ a≼c b≼c) (inl (⊔-idempotent c))
\end{code}

-------------------------------------------------------------------------------

Properties of monus.

\begin{code}
monus-gc-→ : a ∸ p ≼ b → a ≼ b + p
monus-gc-→ {a = a} {p = zero} {b = b} xx = ≼-transitive (inl (symmetric ∸-right-identity)) (≼-transitive xx (inl (symmetric +-right-identity)))
monus-gc-→ {a = zero} {p = succ p} {b = b} xx = ≼-transitive xx (≼-transitive (inl (symmetric +-right-identity)) (k+-preserves-≼ b (inr zero≺succ)))
monus-gc-→ {a = succ a} {p = succ p} {b = b} xx = ≼-transitive (succ≼succ (monus-gc-→ xx)) (inl (shift' b p))
\end{code}

\begin{code}
monus-gc-← : a ∸ p ≼ b ← a ≼ b + p
monus-gc-← {a = a} {p = zero} {b = b} xx = ≼-transitive (inl ∸-right-identity) (≼-transitive xx (inl +-right-identity))
monus-gc-← {a = zero} {p = succ p} {b = b} xx = zero≼n b
monus-gc-← {a = succ a} {p = succ p} {b = b} xx = monus-gc-← (succ-reflects-≼ (≼-transitive xx (inl (shift b p))))
\end{code}

-------------------------------------------------------------------------------

NEW STUFF. TODO: organize!

\begin{code}
-- TODO: MOVE these to VFP.Natural.Order.Properties AND/OR VFP.Natural.Properties!!
genauer : ∀ a → a ≽ 1 → a *2 ≽ 1
genauer zero p = p
genauer (succ a) p = inr (succ≺succ zero≺succ)

genau : ∀ a → 2^ a ≽ 1
genau zero = inl reflexive
genau (succ a) = genauer (2^ a) (genau a)

genau-left : ∀ a → 1 ≡ 1 ⊓ 2^ a
genau-left a = symmetric (connecting-lemma-⊓ (genau a))

genau-right : ∀ a → 1 ≡ 2^ a ⊓ 1
genau-right a = ≡-transitive (genau-left a) (⊓-commutative 1 (2^ a))
\end{code}

\begin{code}
-- ℕ is a lattice and _*2 is a lattice homomorphism
*2-preserves-⊓ : ∀ a b → (a ⊓ b) *2 ≡ a *2 ⊓ b *2
*2-preserves-⊓ zero zero = reflexive
*2-preserves-⊓ zero (succ b) = reflexive
*2-preserves-⊓ (succ a) zero = reflexive
*2-preserves-⊓ (succ a) (succ b) = congruent (λ ◌ → 2 + ◌) (*2-preserves-⊓ a b)

-- ℕ is a lattice and 2^_ is a lattice homomorphism
2^-preserves-⊓ : ∀ a b → 2^ (a ⊓ b) ≡ (2^ a) ⊓ (2^ b)
2^-preserves-⊓ zero zero = reflexive
2^-preserves-⊓ zero (succ b) = genau-left (succ b)
2^-preserves-⊓ (succ a) zero = genau-right (succ a)
2^-preserves-⊓ (succ a) (succ b) = ≡-transitive (congruent _*2 (2^-preserves-⊓ a b)) (*2-preserves-⊓ (2^ a) (2^ b))

two^-preserves-⊓ = 2^-preserves-⊓

⊓-monotone-left : ∀ a₁ a₂ → a₁ ≼ a₂ → ∀ {b} → a₁ ⊓ b ≼ a₂ ⊓ b
⊓-monotone-left zero zero (inl x) = inl reflexive
⊓-monotone-left zero (succ a₂) (inr x) = zero≼n (succ a₂ ⊓ _)
⊓-monotone-left (succ a₁) (succ .a₁) (inl reflexive) = inl reflexive
⊓-monotone-left (succ a₁) (succ a₂) (inr (succ≺succ x)) {b = zero} = inl reflexive
⊓-monotone-left (succ a₁) (succ a₂) (inr (succ≺succ x)) {b = succ b} = succ≼succ (⊓-monotone-left a₁ a₂ (inr x))
-- alternative: connecting-lemma-⊓ᴿ {!!}

⊓-monotone-right : ∀ {a} → ∀ {b₁ b₂} → b₁ ≼ b₂ → a ⊓ b₁ ≼ a ⊓ b₂
⊓-monotone-right p = ≼-transitive (inl (⊓-commutative _ _)) (≼-transitive (⊓-monotone-left _ _ p) (inl (⊓-commutative _ _)))

⊓-monotone : ∀ {a₁ a₂} → a₁ ≼ a₂ → ∀ {b₁ b₂} → b₁ ≼ b₂ → a₁ ⊓ b₁ ≼ a₂ ⊓ b₂
⊓-monotone p q = ≼-transitive (⊓-monotone-left _ _ p) (⊓-monotone-right q)

⊓-preserves-≼ = ⊓-monotone

[a⊓b]*2≼a+b : ∀ a b → (a ⊓ b) *2 ≼ a + b
[a⊓b]*2≼a+b zero zero = inl reflexive
[a⊓b]*2≼a+b zero (succ b) = inr zero≺succ
[a⊓b]*2≼a+b (succ a) zero = inr zero≺succ
[a⊓b]*2≼a+b (succ a) (succ b) = ≼-transitive (succ≼succ (succ≼succ ([a⊓b]*2≼a+b a b))) (inl (congruent succ (shunt-succ a b)))
\end{code}

\begin{code}
*2-preserves-⊔ : ∀ a b → (a ⊔ b) *2 ≡ a *2 ⊔ b *2
*2-preserves-⊔ zero b = reflexive
*2-preserves-⊔ (succ a) zero = reflexive
*2-preserves-⊔ (succ a) (succ b) = congruent (λ hole → succ (succ hole)) (*2-preserves-⊔ a b)

2^-preserves-⊔ : ∀ a b → 2^ (a ⊔ b) ≡ (2^ a) ⊔ (2^ b)
2^-preserves-⊔ zero b = ≡-symmetric (connecting-lemma-⊔ (genau b))
2^-preserves-⊔ (succ a) zero = connecting-lemma-⊔′ (genau (succ a))
2^-preserves-⊔ (succ a) (succ b) = ≡-transitive (congruent _*2 (2^-preserves-⊔ a b)) (*2-preserves-⊔ (2^ a) (2^ b))

-- TODO: move this to where this belongs
n≼n*2 : ∀ n → n ≼ n *2
n≼n*2 zero = inl reflexive
n≼n*2 (succ n) = succ≼succ (≼-transitive (n≼n*2 n) n≼succ-n)

a+b≼[a⊔b]*2 : ∀ a b → a + b ≼ (a ⊔ b) *2
a+b≼[a⊔b]*2 zero b = n≼n*2 b
a+b≼[a⊔b]*2 (succ a) zero rewrite +-right-unit a = succ≼succ (≼-transitive (n≼n*2 a) n≼succ-n)
a+b≼[a⊔b]*2 (succ a) (succ b) = succ≼succ (≼-transitive (inl (shift a b)) (succ≼succ (a+b≼[a⊔b]*2 a b)))
\end{code}

-------------------------------------------------------------------------------

Unicode: ≺ \prec ; ≻ \succ ; ≼ \preceq ; ≽ \succeq

