This file contains supplementary material for the pearl “Binary Search
— Think Positive” such as examples and solutions to exercises.  It is
organised as a series of submodules, basically one for each
weekday and one for each exercise.

-------------------------------------------------------------------------------

\begin{code}
-- no “safe” flag thanks to Harry
{-# OPTIONS --without-K #-}

module VFP.Searching.JFP where

\end{code}

Midpoint-Tree contains the definition of midpoint trees. Since they
are used several times they are in their own module.

\begin{code}
module Midpoint-Tree where
  open import VFP.Natural.Midpoint.Properties-7 public
  open import VFP.Searching.Binary-Search public
  open import VFP.Natural public

  private variable
    l r n : ℕ

  infix 6 _≺′_
\end{code}

<<order-binary-search>>
\begin{code}
  data _≺′_ : ℕ → ℕ → Set where
    one  : n ≺′ 1 + n
    _∔_  : l ≺′ ⌊ l mid r ⌋ → ⌊ l mid r ⌋ ≺′ r → l ≺′ r
\end{code}

-------------------------------------------------------------------------------

Monday:
- The first implementation of even-odd-pair using linear left-to-right
  search strategy.
- Implementation of even-odd-pair using binary search strategy.

\begin{code}
module Monday where
  open import VFP.Logic
  open Midpoint-Tree

  private variable
    l r : ℕ
\end{code}

<<linear-from-left-to-right-with-ordering-7-sig>>
\begin{code}
  even-odd-pair  :  ∀ (l r : ℕ)
                  →  l ≺ r
                  →  (box : ℕ → ℕ)
                  →            even (box l) × odd (box r)
                  →  ∃ (λ i →  even (box i) × odd (box (1 + i)))
\end{code}

<<linear-from-left-to-right-with-ordering-7-impl>>
\begin{code}
  even-odd-pair l r (one)         box (e , o) = l , e , o
  even-odd-pair l r (succ succ-l≺r)  box (e , o) with parity (box (1 + l))
  … | inr  o′  = l , e , o′
  … | inl  e′  = even-odd-pair (1 + l) r succ-l≺r box (e′ , o)
\end{code}

\begin{code}
  even-odd-pair′  :  ∀ (l r : ℕ)
                  →  l ≺′ r
                  →  (box : ℕ → ℕ)
                  →            even (box l) × odd (box r)
                  →  ∃ (λ i →  even (box i) × odd (box (1 + i)))
\end{code}

<<even-odd-binary-search-order>>
\begin{code}
  even-odd-pair′ l r (one)          box (e , o) = l , e , o
  even-odd-pair′ l r (l≺′m ∔ m≺′r)  box (e , o) with parity (box ⌊ l mid r ⌋)
  … | inr  o′  = even-odd-pair′ l ⌊ l mid r ⌋  l≺′m  box (e , o′)
  … | inl  e′  = even-odd-pair′ ⌊ l mid r ⌋ r  m≺′r  box (e′ , o)
\end{code}

-------------------------------------------------------------------------------

Tuesday:
- Binary subdivision scheme.
- The third implementation of even-odd-pair presented in the paper. It
  builds on the new interval tree type [_,_[.

\begin{code}
module Tuesday where
  open import VFP.Searching.Binary-Search
  open import VFP.Logic
  
  private variable
    l r : ℕ
\end{code}

Knuth's comparison tree, see TAOCP 3 p412, Fig. 5.

\begin{code}
  0≺17 : 0 ≺ 17
  0≺17 = zero≺succ

  ex = binary-subdivision _ _ 0≺17
  _ : ex ≡ 
\end{code}

<<example>>
\begin{code}
      Node 8  (Node 4   (Node 2   (Node 1 (Leaf 0) (Leaf 1)) (Node 3 (Leaf 2) (Leaf 3)))
                        (Node 6   (Node 5 (Leaf 4) (Leaf 5)) (Node 7 (Leaf 6) (Leaf 7))))
              (Node 12  (Node 10  (Node 9 (Leaf 8) (Leaf 9)) (Node 11 (Leaf 10) (Leaf 11)))
                        (Node 14  (Node 13 (Leaf 12) (Leaf 13))
                                  (Node 15 (Leaf 14) (Node 16 (Leaf 15) (Leaf 16)))))
\end{code}

\begin{code}
  _ = reflexive
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<even-odd-pair-signature>>
\begin{code}
  even-odd-pair  :  ⟦ l , r ⦆
                 →  (box : ℕ → ℕ) 
                 →            even (box l) × odd (box r)
                 →  ∃ (λ i →  even (box i) × odd (box (1 + i)))
\end{code}
<<even-odd-pair-definition>>
\begin{code}
  even-odd-pair (Leaf n)      box (e , o) = n , e , o
  even-odd-pair (Node m t u)  box (e , o) with parity (box m)
  … | inr  o′  = even-odd-pair t  box (e   , o′  )
  … | inl  e′  = even-odd-pair u  box (e′  , o   )
\end{code}

-------------------------------------------------------------------------------

Wednesday:
- Several subdivision functions:
  - Midpoint-trees to interval trees.
  - Harry's binary subdivision using a terminating pragma.
  - Binary subdivision using accessibility in the worker/wrapper scheme.

\begin{code}
module Wednesday where
  open Midpoint-Tree

  private variable
    l r : ℕ
\end{code}

<<binary-subdivision>>
\begin{code}
  binary-subdivision″ : l ≺′ r -> ⟦ l , r ⦆
  binary-subdivision″ (one)    = Leaf _
  binary-subdivision″ (t ∔ u)  = Node _ (binary-subdivision″ t) (binary-subdivision″ u)
\end{code}

Constructing an element of l ≺′ r:

\begin{code}
  {-# TERMINATING #-}
\end{code}

<<Harrys>>
\begin{code}
  balanced′ : l ≺ r -> l ≺′ r
  balanced′ (one)     = one
  balanced′ (succ p)  = balanced′ (a≺a-mid-b p) ∔ balanced′  (a-mid-b≺b (succ p))
\end{code}

<<Kens-worker>>
\begin{code}
  worker : l ≺ r -> Accessible (r ∸ l) -> l ≺′ r
  worker (one)     (access because) = one
  worker (succ p)  (access because) =
       worker  (a≺a-mid-b p)         (because (shrink-left (succ p)))
    ∔  worker  (a-mid-b≺b (succ p))  (because (shrink-right p))
\end{code}

<<Kens-wrapper>>
\begin{code}
  balanced : l ≺ r → l ≺′ r
  balanced l≺r = worker l≺r accessible
\end{code}

-------------------------------------------------------------------------------

Thursday:
- Binary search using interval-search.

\begin{code}
module Thursday where
  open import VFP.Searching.Binary-Search
  open import VFP.Logic
  open Wednesday

  private variable
    l r : ℕ

  variable
    P Q : ℕ → Set
\end{code}

<<binary-search>>
\begin{code}
  binary-search′  : l ≺ r
                  → (oracle : (n : ℕ) -> P n ⊎ Q n)
                  →            P l × Q r
                  → ∃ (λ i ->  P i × Q (1 + i))
  binary-search′ l≺r = interval-search (binary-subdivision″ (balanced l≺r))
\end{code}

-------------------------------------------------------------------------------

Friday:
- We show that the returned indices are indeed in the given interval
  for ord-table-lookup and find.

\begin{code}
module Friday where
  open import VFP.Searching.Binary-Search.Applications
  open import VFP.Natural

  private variable
    l r : ℕ

  variable
    P Q : ℕ → Set
\end{code}

\begin{code}
  ord-table-lookup+  : l ≺ r
                    → (table : ℕ → ℕ) 
                    → (monotone : ∀ {a b} → a ≼ b → table a ≼ table b)
                    → (key : ℕ)
                    → table l ≼ key × key ≺ table r
                    → Decide (∃ (λ i → l ≼ i × i ≺ r × table i ≡ key))
  ord-table-lookup+                 l≺r table monotone key pre                        with table-lookup l≺r table key pre
  ord-table-lookup+ {l = l} {r = r} l≺r table monotone key pre                             | i , inl pᵢ , qᵢ   with l ≼?≻ i | i ≺?≽ r
  ord-table-lookup+                 l≺r table monotone key pre                             | i , inl pᵢ , qᵢ        |  inl l≼i | inl i≺r   = yes (i , (l≼i , (i≺r , pᵢ)))
  ord-table-lookup+ {l = l}         l≺r table monotone key (inl table-l≡key , key≺table-r) | i , inl pᵢ , qᵢ        |  inr l≻i | inl i≺r   = yes (l , (inl reflexive) , (l≺r , table-l≡key))
  ord-table-lookup+                 l≺r table monotone key (table-l≼key , key≺table-r)     | i , inl pᵢ , qᵢ        |  inl l≼i | inr i≽r rewrite (≡-symmetric pᵢ)
    = ex-falso-quodlibet (≺-irreflexive (≼-then-≺ (monotone i≽r) key≺table-r))
  ord-table-lookup+                 l≺r table monotone key (inr table-l≺key , key≺table-r) | i , inl pᵢ , qᵢ        |  inr l≻i | inl i≺r rewrite (≡-symmetric pᵢ)
    = ex-falso-quodlibet (≺-irreflexive (≼-then-≺ (monotone (inr l≻i)) table-l≺key))
  ord-table-lookup+                 l≺r table monotone key pre                             | i , inl pᵢ , qᵢ        |  inr l≻i | inr i≽r
    = ex-falso-quodlibet (≺-irreflexive (≺-transitive l≻i (≺-then-≼ l≺r i≽r)))
  ord-table-lookup+                 l≺r table monotone key pre                             | i , inr pᵢ , qᵢ
    = no key-not-found
      where  key-not-found : ¬ ∃ (λ j → l ≼ j × j ≺ r × table j ≡ key)
             key-not-found (j , _ , _ , reflexive) with j ≼?≻ i
             ... | inl  j≼i  = ≺-irreflexive (≺-then-≼ pᵢ (monotone j≼i))
             ... | inr  j≻i  = ≺-irreflexive (≺-then-≼ qᵢ (monotone (≺-is-succ-≼ j≻i)))
\end{code}

\begin{code}
  find+  : l ≺ r
        → (monotone : ∀ {a b} → a ≼ b → (P a → P b))
        → (decide : ∀ n -> ¬ P n ⊎ P n)
        → ¬ P l × P r
        → ∃ (λ j ->  l ≺ j × j ≼ r × P j × ∀ i → i ≺ j → ¬ P i)
  find+                 l≺r monotone decide pre with binary-search l≺r decide pre
  find+ {l = l} {r = r} l≺r monotone decide pre | j , ¬P-j , P-1+j with l ≺?≽ (1 + j) | (1 + j) ≼?≻ r
  find+                 l≺r monotone decide pre | j , ¬P-j , P-1+j      | inl l≺1+j      | inl 1+j≼r
    = (1 + j) , l≺1+j , 1+j≼r , P-1+j , λ i (i≺1+j : i ≺ 1 + j) P-i → ¬P-j (monotone (≺-succ-is-≼ i≺1+j) P-i)
  find+                 l≺r monotone decide pre | j , ¬P-j , P-1+j      | inl l≺1+j      | inr 1+j≻r
    = ex-falso-quodlibet (¬P-j (monotone (≺-succ-is-≼ 1+j≻r) (outr pre)))
  find+                 l≺r monotone decide pre | j , ¬P-j , P-1+j      | inr l≽1+j | _
    = ex-falso-quodlibet (outl pre (monotone l≽1+j P-1+j))
\end{code}

-------------------------------------------------------------------------------

Exercise 1

\begin{code}
module Exercise-1 where
  open Midpoint-Tree

  private variable
    l r : ℕ
\end{code}

<<Harrys-forward-direction-sig>>
\begin{code}
  domain′ : l ≺′ r → l ≺ r
\end{code}
\begin{code}
  domain′ (one)   = one
  domain′ (p ∔ q) = ≺-transitive (domain′ p) (domain′ q)
\end{code}

-------------------------------------------------------------------------------

Exercises 2 and 3:
- Define membership in closed intervals.
- n ⋵ I is equivalent to l ≼ n × n ≼ r for I : ⟦ l , r ⦆.
- interval-search for predicates and oracles over finite domains.

\begin{code}
module Exercise-2-3 where
  open import VFP.Searching.Binary-Search hiding (interval-search+ ; ↙_ )
  open import VFP.Natural
  open import VFP.Logic

  private variable
    i l r : ℕ

  infix 6 _⋵_
\end{code}

<<closed-membership-definition>>
\begin{code}
  data _⋵_ (i : ℕ) : ⟦ l , r ⦆ → Set where
    half-open  : {range : ⟦ l , r  ⦆} → i ∈ range → i ⋵ range
    rightmost  : {range : ⟦ l , i  ⦆} → i ⋵ range
\end{code}

<<left-right-in-range-signature>>
\begin{code}
  ↙_  : (range : ⟦ l , r ⦆) → l  ⋵ range
  _↘  : (range : ⟦ l , r ⦆) → r  ⋵ range
\end{code}

\begin{code}
  ↙ range = half-open (leftmost range)
  
  range ↘ = rightmost
\end{code}

\begin{code}
  closed-bounds : ∀ {t : ⟦ l , r ⦆} -> i ⋵ t -> l ≼ i × i ≼ r
  closed-bounds (half-open x) = let (l≼i , i≺r) = bounds x in l≼i , (inr i≺r)
  closed-bounds {t = t} rightmost = (inr (domain t)) , inl reflexive

  closed-contains : ∀ (t : ⟦ l , r ⦆) -> l ≼ i × i ≼ r -> i ⋵ t
  closed-contains t (l≼i , inl reflexive) = rightmost
  closed-contains t (l≼i , inr i≺r) = half-open (contains t (l≼i , i≺r))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
  middle-member : {m : ℕ} → (t : ⟦ l , m ⦆) → (u : ⟦ m , r ⦆) → m ∈ Node m t u
  middle-member _ u = right (leftmost u)

  closed-middle-member : {m : ℕ} → (t : ⟦ l , m ⦆) → (u : ⟦ m , r ⦆) → m ⋵ Node m t u
  closed-middle-member t u = half-open (middle-member t u)

  closed-left : {n m : ℕ} → {t : ⟦ l , m ⦆} → {u : ⟦ m , r ⦆} → n ⋵ t → n ⋵ Node m t u
  closed-left (half-open n∈t) = half-open (left n∈t)
  closed-left rightmost = closed-middle-member _ _

  closed-right : {n m : ℕ} → {t : ⟦ l , m ⦆} → {u : ⟦ m , r ⦆} → n ⋵ u → n ⋵ Node m t u
  closed-right (half-open n∈u) = half-open (right n∈u)
  closed-right rightmost = rightmost
\end{code}

<<interval-search++-signature>>
\begin{code}
  interval-searchꟳ  : (range : ⟦ l , r ⦆)
                    → {P Q :  (n : ℕ) → n ⋵ range → Set}
                    → (oracle : (n : ℕ) -> (p : n ⋵ range) → P n p ⊎ Q n p)
                    → P l (↙ range) × Q r (range ↘)
                    → ∃ (λ i → ∃ (λ p → ∃ (λ q → P i p × Q (1 + i) q)))
\end{code}
\begin{code}
  interval-searchꟳ (Leaf l)     oracle (p , q) = l , half-open here , rightmost , p , q
  interval-searchꟳ (Node m t u) oracle (p , q) with oracle m (closed-middle-member t u)
  … | inl p′ = let (i , i∈r , 1+i∈r , pi , qi) = interval-searchꟳ u (λ n n∈u → oracle n (closed-right n∈u)) (p′ , q) in i , closed-right i∈r , closed-right 1+i∈r , pi , qi
  … | inr q′ = let (i , i∈r , 1+i∈r , pi , qi) = interval-searchꟳ t (λ n n∈u → oracle n (closed-left  n∈u)) (p , q′) in i , closed-left  i∈r , closed-left  1+i∈r , pi , qi
\end{code}

-------------------------------------------------------------------------------

Exercise 4:
- Implementation of interval-search over the integers.

\begin{code}
module Exercise-4 where
  open import VFP.Integer
  open import VFP.Integer.Properties
  open import VFP.Logic

  private variable
    l r : ℤ
  
  variable
    P Q : ℤ -> Set
\end{code}

\begin{code}
  data ⟦_,_⦆ : ℤ → ℤ → Set where
    Leaf  : (n  : ℤ) -> ⟦ n , succ n ⦆
    Node  : (m  : ℤ) -> ⟦ l , m ⦆ -> ⟦ m , r ⦆ -> ⟦ l , r ⦆
\end{code}

\begin{code}
  interval-search  : ⟦ l , r ⦆
                   → (oracle : ∀ n -> P n ⊎ Q n)
                   →            P l × Q r
                   → ∃ (λ i ->  P i × Q (succ i))
  interval-search (Leaf n)       oracle (p , q) = n , p , q
  interval-search (Node m t u)   oracle (p , q) with oracle m
  ... | inl  p′  = interval-search u  oracle (p′  , q   )
  ... | inr  q′  = interval-search t  oracle (p   , q′  )
\end{code}

-------------------------------------------------------------------------------
